"""
Shared utils for dealing with XML.
"""

import glob
import re
import os.path
from xml.etree import ElementTree

def get_xml_element(xml_file, element_name):
    """ Gets a given element from a given xml file.

    Raises
    ------
    ValueError
        If the given xml_file does not exist.
        If the given xml_file does not contain an element with the given element_name.

    Returns
    -------
    xml.etree.ElementTree.Element
        The Element matching the given element_name.
    """

    # verify runtime config
    if not os.path.exists(xml_file):
        raise ValueError('Given xml file does not exist: ' + xml_file)

    # parse the xml file and figure out the namespace if there is one
    xml_xml = ElementTree.parse(xml_file)
    xml_root = xml_xml.getroot()
    xml_namespace_match = re.match(r'\{.*}', str(xml_root.tag))
    xml_namespace = ''
    if xml_namespace_match:
        xml_namespace = xml_namespace_match.group(0)

    # extract needed information from the xml file
    if xml_root.tag == element_name:
        xml_element = xml_root
    else:
        xml_element = xml_root.find('./' + xml_namespace + element_name)

    # verify information from xml file
    if xml_element is None:
        raise ValueError('Given xml file (' + \
             xml_file + \
             ') does not have ./' + \
             element_name + \
             ' element' \
        )

    return xml_element

def get_xml_element_by_path(
    xml_file_path,
    xpath,
    default_namespace=None,
    xml_namespace_dict=None,
    find_all=False
):
    """Gets the XML element(s) from a given xml file given an xpath.

    Parameters
    ----------
    xml_file_path : str
        Path of the xml file
    xpath : str
        Xpath of the element you want
    default_namespace : str
        Optional string specifying the default namespace you are using in your xpath selector.
        This is the most common argument that will most likely be used.
        If your XML is namespaced, then even if your elements are in the default namespace,
        you must specify and use this namespace in both your xpath as well as specifying it here.
    xml_namespace_dict : Dict[str, str]
        Optional dictionary if default_namespace is not enough and you have multiple
        namespaces that you need to deal with in your xpath selector.
    find_all : bool
        If False find only the first matching Element.
        If True find all matching elements.

    Returns
    -------
    xml.etree.ElementTree.Element or [xml.etree.ElementTree.Element]
        The Element(s) found given the xpath
    """
    # verify runtime config
    if not os.path.exists(xml_file_path):
        raise ValueError(f'Given xml file does not exist: {xml_file_path}')

    # figure out the xml namespaceing
    xml_file = ElementTree.parse(xml_file_path).getroot()
    namespaces = xml_namespace_dict
    if xml_namespace_dict is None and default_namespace is not None:
        xml_namespace_match = re.findall(r'{(.*?)}', xml_file.tag)
        xml_namespace = xml_namespace_match[0] if xml_namespace_match else ''
        namespaces = {}
        namespaces[default_namespace] = xml_namespace

    # find the element(s)
    results = None
    if find_all:
        results = xml_file.findall(xpath, namespaces)
    else:
        results = xml_file.find(xpath, namespaces)

    return results

def get_xml_element_text_by_path(
    xml_file_path,
    xpath,
    default_namespace=None,
    xml_namespace_dict=None,
    find_all=False
):
    """Gets the text of XML element(s) from a given xml file given an xpath.

    Parameters
    ----------
    xml_file_path : str
        Path of the xml file
    xpath : str
        Xpath of the element you want
    default_namespace : str
        Optional string specifying the default namespace you are using in your xpath selector.
        This is the most common argument that will most likely be used.
        If your XML is namespaced, then even if your elements are in the default namespace,
        you must specify and use this namespace in both your xpath as well as specifying it here.
    xml_namespace_dict : Dict[str, str]
        Optional dictionary if default_namespace is not enough and you have multiple
        namespaces that you need to deal with in your xpath selector.
    find_all : bool
        If False find only the first matching Element.
        If True find all matching elements.

    Returns
    -------
    str or [str]
        The text of the XML elements found given the xpath
    """
    xml_elements = get_xml_element_by_path(
        xml_file_path=xml_file_path,
        xpath=xpath,
        default_namespace=default_namespace,
        xml_namespace_dict=xml_namespace_dict,
        find_all=find_all
    )

    result = None
    if find_all:
        result = []
        for xml_element in xml_elements:
            result.append(xml_element.text)
    else:
        result = xml_elements.text

    return result

def aggregate_xml_element_attribute_values(xml_file_paths, xml_element, attribs): # pylint: disable=too-many-branches
    """Function to parse a XML file looking for a specific element and obtaining its attributes.

    If element is not found in file the file is ignored.

    If attribute is not found in element then any other attributes that are in the file are still
    aggregated.

    Parameters
    ----------
    xml_file_paths : str, [str]
        Path(s) to XML file(s) or directory(s) containing XML file(s)
    xml_element: str
        XML element being searched for
    attribs: list
        List of attributes being searched for in the XML element

    Returns
    --------
    report_results: dict
        A dictionary of attribute values.
    """
    report_results = {}

    if not isinstance(xml_file_paths, list):
        xml_file_paths = [xml_file_paths]

    # collect all the xml file paths
    xml_files = []
    for xml_file_path in xml_file_paths:
        if os.path.isdir(xml_file_path):
            xml_files += glob.glob(xml_file_path + '/*.xml', recursive=False)
        elif os.path.isfile(xml_file_path):
            xml_files += [xml_file_path]

    #Iterate over XML files
    for file in xml_files:
        try:
            #Add attributes to dictionary
            element = get_xml_element(file, xml_element)
            for attrib in attribs:
                if attrib in element.attrib:
                    # aggregate attribute accross multiple files
                    attribute_value = element.attrib[attrib]

                    if not attrib in report_results:
                        report_results[attrib] = []
                    report_results[attrib].append(attribute_value)
                else:
                    # ignore if a file has the element but missing an attribute
                    print(
                        f"WARNING: Did not find attribute ({attrib}) on element ({element})" \
                        f" in file ({file}). Ignoring."
                    )
                    continue
        except ValueError:
            # ignore if a file has the element but missing an attribute
            print(
                f"WARNING: Did not find element ({xml_element})" \
                f" in file ({file}). Ignoring."
            )
            continue

    #Loop to check if all values aggregated are integers, floats, strings, or a mix
    #of all three.
    for attrib in attribs:
        only_numerics = False
        only_floats = False

        #Iteration to determine what pattern all values follow.
        if attrib in report_results:
            for value in report_results[attrib]:
                #isnumeric() returns True if the value contains numbers
                #but not a decimal point. If this returns
                #true then continue because the value would be
                #a valid float anyway
                if value.isnumeric():
                    only_numerics = True
                    only_floats = True
                else:
                    only_numerics = False
                    try:
                        float(value)
                        only_floats = True
                    except ValueError:
                        only_floats = False
                        #Break since only_numerics and only_floats are both false
                        break

            if only_numerics or only_floats:
                aggregated_value = 0
            else:
                aggregated_value = []

            #Iteration to aggregate values based on determined pattern.
            for value in report_results[attrib]:
                #If a value is numeric and can be a float then it is an integer
                if only_numerics:
                    aggregated_value = int(aggregated_value) + int(value)
                #If a value is not numeric but can be a float then it is a float
                elif only_floats:
                    aggregated_value = float(aggregated_value) + float(value)
                else:
                    aggregated_value.append(value)
        else:
            aggregated_value = None

        #Replace list of values with aggregated value that is either an int,
        #float, or string
        if aggregated_value is not None:
            report_results[attrib] = aggregated_value

    return report_results



import os

from testfixtures import TempDirectory
from tests.helpers.base_step_implementer_test_case import \
    BaseStepImplementerTestCase
from ploigos_step_runner.step_implementers.generate_metadata import Npm
from ploigos_step_runner import StepResult


class TestStepImplementerGenerateMetadataNpm(BaseStepImplementerTestCase):
    def create_step_implementer(
            self,
            step_config={},
            step_name='',
            implementer='',
            workflow_result=None,
            parent_work_dir_path=''
    ):
        return self.create_given_step_implementer(
            step_implementer=Npm,
            step_config=step_config,
            step_name=step_name,
            implementer=implementer,
            workflow_result=workflow_result,
            parent_work_dir_path=parent_work_dir_path
        )

    def test_step_implementer_config_defaults(self):
        defaults = Npm.step_implementer_config_defaults()
        expected_defaults = {
            'package-file': 'package.json'
        }
        self.assertEqual(defaults, expected_defaults)

    def test__required_config_or_result_keys(self):
        required_keys = Npm._required_config_or_result_keys()
        expected_required_keys = [
            'package-file'
        ]
        self.assertEqual(required_keys, expected_required_keys)

    def test__validate_required_config_or_previous_step_result_artifact_keys_valid(self):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')

            temp_dir.write('package.json', b'''{
              "name": "my-awesome-package",
              "version": "42.1"
            }''')
            package_file_path = os.path.join(temp_dir.path, 'package.json')

            step_config = {
                'package-file': package_file_path
            }

            step_implementer = self.create_step_implementer(
                step_config=step_config,
                step_name='test',
                implementer='Npm',
                parent_work_dir_path=parent_work_dir_path
            )

            step_implementer._validate_required_config_or_previous_step_result_artifact_keys()

    def test__validate_required_config_or_previous_step_result_artifact_keys_package_file_does_not_exist(self):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            package_file_path = os.path.join(temp_dir.path, 'package.json')
            step_config = {
                'package-file': package_file_path
            }

            step_implementer = self.create_step_implementer(
                step_config=step_config,
                step_name='test',
                implementer='Npm',
                parent_work_dir_path=parent_work_dir_path
            )

            with self.assertRaisesRegex(
                AssertionError,
                rf"Given npm package file \(package-file\) does not exist: {package_file_path}"
            ):
                step_implementer._validate_required_config_or_previous_step_result_artifact_keys()

    def test_run_step_pass(self):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            temp_dir.write('package.json', b'''{
              "name": "my-awesome-package",
              "version": "42.1"
            }''')
            package_file_path = os.path.join(temp_dir.path, 'package.json')

            step_config = {'package-file': package_file_path}

            step_implementer = self.create_step_implementer(
                step_config=step_config,
                step_name='generate-metadata',
                implementer='Npm',
                parent_work_dir_path=parent_work_dir_path,
            )

            result = step_implementer._run_step()

            expected_step_result = StepResult(
                step_name='generate-metadata',
                sub_step_name='Npm',
                sub_step_implementer_name='Npm'
            )
            expected_step_result.add_artifact(name='app-version', value='42.1')

            self.assertEqual(result, expected_step_result)


    def test_run_step_fail_missing_version_in_package_file(self):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            temp_dir.write('package.json', b'''{
              "name": "my-awesome-package"
            }''')
            package_file_path = os.path.join(temp_dir.path, 'package.json')

            step_config = {'package-file': package_file_path}

            step_implementer = self.create_step_implementer(
                step_config=step_config,
                step_name='generate-metadata',
                implementer='Npm',
                parent_work_dir_path=parent_work_dir_path,
            )

            result = step_implementer._run_step()

            expected_step_result = StepResult(
                step_name='generate-metadata',
                sub_step_name='Npm',
                sub_step_implementer_name='Npm',
            )
            expected_step_result.success = False
            expected_step_result.message = f"Given npm package file ({package_file_path})" + \
                ' does not contain a \"version\" key.'

            self.assertEqual(result, expected_step_result)


from ploigos_step_runner import StepImplementer, StepResult
from ploigos_step_runner.config.config_value import ConfigValue


class FailStepImplementer(StepImplementer):
    @staticmethod
    def step_implementer_config_defaults():
        return {}

    @staticmethod
    def _required_config_or_result_keys():
        return []

    def _run_step(self):
        step_result = StepResult.from_step_implementer(self)
        step_result.success = False
        return step_result


class FooStepImplementer(StepImplementer):
    @staticmethod
    def step_implementer_config_defaults():
        return {}

    @staticmethod
    def _required_config_or_result_keys():
        return []

    def _run_step(self):
        step_result = StepResult.from_step_implementer(self)
        return step_result

class FooStepImplementer2(StepImplementer):
    @staticmethod
    def step_implementer_config_defaults():
        return {}

    @staticmethod
    def _required_config_or_result_keys():
        return []

    def _run_step(self):
        step_result = StepResult.from_step_implementer(self)
        return step_result


class RequiredStepConfigStepImplementer(StepImplementer):
    @staticmethod
    def step_implementer_config_defaults():
        return {}

    @staticmethod
    def _required_config_or_result_keys():
        return [
            'required-config-key'
        ]

    def _run_step(self):
        step_result = StepResult.from_step_implementer(self)
        runtime_step_config = self.config.get_copy_of_runtime_step_config(
            self.environment,
            self.step_implementer_config_defaults())
        for n, v in ConfigValue.convert_leaves_to_values(runtime_step_config).items():
            step_result.add_artifact(name=n, value=v)
        return step_result

class RequiredStepConfigMultipleOptionsStepImplementer(StepImplementer):
    @staticmethod
    def step_implementer_config_defaults():
        return {}

    @staticmethod
    def _required_config_or_result_keys():
        return [
            ['new-required-key', 'old-required-key']
        ]

    def _run_step(self):
        step_result = StepResult.from_step_implementer(self)
        runtime_step_config = self.config.get_copy_of_runtime_step_config(
            self.environment,
            self.step_implementer_config_defaults())
        for n, v in ConfigValue.convert_leaves_to_values(runtime_step_config).items():
            step_result.add_artifact(name=n, value=v)
        return step_result
class WriteConfigAsResultsStepImplementer(StepImplementer):
    @staticmethod
    def step_implementer_config_defaults():
        return {}

    @staticmethod
    def _required_config_or_result_keys():
        return []

    def _run_step(self):
        step_result = StepResult.from_step_implementer(self)
        runtime_step_config = self.config.get_copy_of_runtime_step_config(
            self.environment,
            self.step_implementer_config_defaults())

        # copy the key/value pairs into the artifacts
        for name, value in ConfigValue.convert_leaves_to_values(runtime_step_config).items():
            # print(name, value)
            step_result.add_artifact(name, value)
        return step_result


class NotSubClassOfStepImplementer():
    pass



import os
import zipfile
from pathlib import Path
from unittest.mock import patch

import mock
from ploigos_step_runner import StepResult
from ploigos_step_runner.results.workflow_result import WorkflowResult
from ploigos_step_runner.step_implementers.report import ResultArtifactsArchive
from testfixtures import TempDirectory
from tests.helpers.base_step_implementer_test_case import \
    BaseStepImplementerTestCase


class TestStepImplementerResultArtifactsArchiveBase(BaseStepImplementerTestCase):
    def create_step_implementer(
            self,
            step_config={},
            workflow_result=None,
            parent_work_dir_path=''
    ):
        return self.create_given_step_implementer(
            step_implementer=ResultArtifactsArchive,
            step_config=step_config,
            step_name='report',
            implementer='ResultArtifactsArchive',
            workflow_result=workflow_result,
            parent_work_dir_path=parent_work_dir_path
        )

class TestStepImplementerResultArtifactsArchive_other(TestStepImplementerResultArtifactsArchiveBase):
    def test_step_implementer_config_defaults(self):
        defaults = ResultArtifactsArchive.step_implementer_config_defaults()
        expected_defaults = {
            'results-archive-format': 'zip',
            'results-archive-ignored-artifacts': [
                'package-artifacts',
                'image-tar-file'
            ]
        }
        self.assertEqual(defaults, expected_defaults)

    def test__required_config_or_result_keys(self):
        required_keys = ResultArtifactsArchive._required_config_or_result_keys()
        expected_required_keys = [
            'organization',
            'application-name',
            'service-name',
            'version',
            'results-archive-format'
        ]
        self.assertEqual(required_keys, expected_required_keys)

class TestStepImplementerResultArtifactsArchive_run_step(TestStepImplementerResultArtifactsArchiveBase):
    @patch.object(
        ResultArtifactsArchive,
        '_ResultArtifactsArchive__create_archive',
        return_value=None
    )
    def test__run_step_pass_no_archive(self, create_archive_mock):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            step_config = {
                'organization': 'test-ORG',
                'application-name': 'test-APP',
                'service-name': 'test-SERVICE',
                'version': '42.0-test'
            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
            )

            step_result = step_implementer._run_step()
            expected_step_result = StepResult(
                step_name='report',
                sub_step_name='ResultArtifactsArchive',
                sub_step_implementer_name='ResultArtifactsArchive'
            )
            expected_step_result.add_artifact(
                name='result-artifacts-archive',
                value='No result artifact values to archive.',
                description='Archive of all of the step result artifacts marked for archiving.'
            )
            self.assertEqual(step_result, expected_step_result)

            create_archive_mock.assert_called_once()

    @patch.object(
        ResultArtifactsArchive,
        '_ResultArtifactsArchive__create_archive',
        return_value='/fake/archive/path/foo.zip'
    )
    def test__run_step_pass(self, create_archive_mock):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            step_config = {
                'organization': 'test-ORG',
                'application-name': 'test-APP',
                'service-name': 'test-SERVICE',
                'version': '42.0-test'
            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
            )

            step_result = step_implementer._run_step()
            expected_step_result = StepResult(
                step_name='report',
                sub_step_name='ResultArtifactsArchive',
                sub_step_implementer_name='ResultArtifactsArchive'
            )
            expected_step_result.add_artifact(
                name='result-artifacts-archive',
                value='/fake/archive/path/foo.zip',
                description='Archive of all of the step result artifacts marked for archiving.'
            )
            self.assertEqual(step_result, expected_step_result)

            create_archive_mock.assert_called_once()

    @patch('ploigos_step_runner.step_implementers.report.result_artifacts_archive.upload_file')
    @patch.object(
        ResultArtifactsArchive,
        '_ResultArtifactsArchive__create_archive',
        return_value='/fake/archive/path/foo.zip'
    )
    def test__run_step_pass_upload_to_file(self, create_archive_mock, upload_file_mock):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            step_config = {
                'organization': 'test-ORG',
                'application-name': 'test-APP',
                'service-name': 'test-SERVICE',
                'version': '42.0-test',
                'results-archive-destination-url': '/mock/results-archives'
            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
            )

            # mock the upload results
            upload_file_mock.return_value = "mock upload results"

            # run the step
            step_result = step_implementer._run_step()

            # verify results
            expected_step_result = StepResult(
                step_name='report',
                sub_step_name='ResultArtifactsArchive',
                sub_step_implementer_name='ResultArtifactsArchive'
            )
            expected_step_result.add_artifact(
                name='result-artifacts-archive',
                value='/fake/archive/path/foo.zip',
                description='Archive of all of the step result artifacts marked for archiving.'
            )
            expected_step_result.add_artifact(
                name='results-archive-uri',
                description='URI of the uploaded results archive.',
                value='/mock/results-archives/test-ORG/test-APP/test-SERVICE/foo.zip'
            )
            expected_step_result.add_artifact(
                name='results-archive-upload-results',
                description='Results of uploading the results archive to the given destination.',
                value='mock upload results'
            )
            self.assertEqual(step_result, expected_step_result)

            # verify mocks called
            create_archive_mock.assert_called_once()
            upload_file_mock.assert_called_once_with(
                file_path=mock.ANY,
                destination_uri='/mock/results-archives/test-ORG/test-APP/test-SERVICE/foo.zip',
                username=None,
                password=None
            )

    @patch('ploigos_step_runner.step_implementers.report.result_artifacts_archive.upload_file')
    @patch.object(
        ResultArtifactsArchive,
        '_ResultArtifactsArchive__create_archive',
        return_value='/fake/archive/path/foo.zip'
    )
    def test__run_step_pass_upload_to_remote_with_auth(self, create_archive_mock, upload_file_mock):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            step_config = {
                'organization': 'test-ORG',
                'application-name': 'test-APP',
                'service-name': 'test-SERVICE',
                'version': '42.0-test',
                'results-archive-destination-url': 'https://ploigos.com/mock/results-archives',
                'results-archive-destination-username': 'test-user',
                'results-archive-destination-password': 'test-pass'
            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
            )

            # mock the upload results
            upload_file_mock.return_value = "mock upload results"

            # run the step
            step_result = step_implementer._run_step()

            # verify results
            expected_step_result = StepResult(
                step_name='report',
                sub_step_name='ResultArtifactsArchive',
                sub_step_implementer_name='ResultArtifactsArchive'
            )
            expected_step_result.add_artifact(
                name='result-artifacts-archive',
                value='/fake/archive/path/foo.zip',
                description='Archive of all of the step result artifacts marked for archiving.'
            )
            expected_step_result.add_artifact(
                name='results-archive-uri',
                description='URI of the uploaded results archive.',
                value='https://ploigos.com/mock/results-archives/test-ORG/test-APP/test-SERVICE/foo.zip'
            )
            expected_step_result.add_artifact(
                name='results-archive-upload-results',
                description='Results of uploading the results archive to the given destination.',
                value='mock upload results'
            )
            self.assertEqual(step_result, expected_step_result)

            # verify mocks called
            create_archive_mock.assert_called_once()
            upload_file_mock.assert_called_once_with(
                file_path=mock.ANY,
                destination_uri='https://ploigos.com/mock/results-archives/test-ORG/test-APP/test-SERVICE/foo.zip',
                username='test-user',
                password='test-pass'
            )

    @patch('ploigos_step_runner.step_implementers.report.result_artifacts_archive.upload_file')
    @patch.object(
        ResultArtifactsArchive,
        '_ResultArtifactsArchive__create_archive',
        return_value='/fake/archive/path/foo.zip'
    )
    def test__run_step_pass_upload_to_remote_with_auth_failure(self, create_archive_mock, upload_file_mock):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            step_config = {
                'organization': 'test-ORG',
                'application-name': 'test-APP',
                'service-name': 'test-SERVICE',
                'version': '42.0-test',
                'results-archive-destination-url': 'https://ploigos.com/mock/results-archives',
                'results-archive-destination-username': 'test-user',
                'results-archive-destination-password': 'test-pass'
            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
            )

            # mock the upload results
            upload_file_mock.side_effect = RuntimeError('mock upload error')

            # run the step
            step_result = step_implementer._run_step()

            # verify results
            expected_step_result = StepResult(
                step_name='report',
                sub_step_name='ResultArtifactsArchive',
                sub_step_implementer_name='ResultArtifactsArchive'
            )
            expected_step_result.add_artifact(
                name='result-artifacts-archive',
                value='/fake/archive/path/foo.zip',
                description='Archive of all of the step result artifacts marked for archiving.'
            )
            expected_step_result.add_artifact(
                name='results-archive-uri',
                description='URI of the uploaded results archive.',
                value='https://ploigos.com/mock/results-archives/test-ORG/test-APP/test-SERVICE/foo.zip'
            )
            expected_step_result.success = False
            expected_step_result.message = 'mock upload error'

            # verify mocks called
            create_archive_mock.assert_called_once()
            upload_file_mock.assert_called_once_with(
                file_path=mock.ANY,
                destination_uri='https://ploigos.com/mock/results-archives/test-ORG/test-APP/test-SERVICE/foo.zip',
                username='test-user',
                password='test-pass'
            )


class TestStepImplementerResultArtifactsArchive__create_archive(TestStepImplementerResultArtifactsArchiveBase):
    def test_no_results(self):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            step_config = {
                'organization': 'test-ORG',
                'application-name': 'test-APP',
                'service-name': 'test-SERVICE',
                'version': '42.0-test'
            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
            )

            archive_path = step_implementer._ResultArtifactsArchive__create_archive()

            self.assertIsNone(archive_path)

    def test_string_result(self):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            step_config = {
                'organization': 'test-ORG',
                'application-name': 'test-APP',
                'service-name': 'test-SERVICE',
                'version': '42.0-test'
            }
            step_result = StepResult(
                step_name='test-step',
                sub_step_name='test-sub-step',
                sub_step_implementer_name='test-sub-step-implementer'
            )
            step_result.add_artifact(
                name='test-step-result-str',
                value='hello world'
            )
            workflow_result = WorkflowResult()
            workflow_result.add_step_result(step_result)
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
                workflow_result=workflow_result
            )

            archive_path = step_implementer._ResultArtifactsArchive__create_archive()

            archive_zip = zipfile.ZipFile(archive_path)

            artifact_file_path = f"{step_config['organization']}-" \
                f"{step_config['application-name']}-{step_config['service-name']}-" \
                f"{step_config['version']}/test-step/test-sub-step/test-step-result-str"
            with archive_zip.open(artifact_file_path, 'r') as artifact_file:
                artifact_file_contents = artifact_file.read().decode('UTF-8')

                self.assertEqual(artifact_file_contents, 'hello world')

    def test_file_result(self):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            step_config = {
                'organization': 'test-ORG',
                'application-name': 'test-APP',
                'service-name': 'test-SERVICE',
                'version': '42.0-test'
            }

            artifact_file_name = 'test-artifact.txt'
            temp_dir.write(artifact_file_name, bytes('hello world file contents', 'utf-8'))

            step_result = StepResult(
                step_name='test-step',
                sub_step_name='test-sub-step',
                sub_step_implementer_name='test-sub-step-implementer'
            )
            step_result.add_artifact(
                name='test-step-result-str',
                value=os.path.join(temp_dir.path, artifact_file_name)
            )
            workflow_result = WorkflowResult()
            workflow_result.add_step_result(step_result)


            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
                workflow_result=workflow_result
            )

            archive_path = step_implementer._ResultArtifactsArchive__create_archive()

            archive_zip = zipfile.ZipFile(archive_path)

            artifact_file_path = f"{step_config['organization']}-" \
                f"{step_config['application-name']}-{step_config['service-name']}-" \
                f"{step_config['version']}/test-step/test-sub-step/test-step-result-str/" \
                f"{artifact_file_name}"
            with archive_zip.open(artifact_file_path, 'r') as artifact_file:
                artifact_file_contents = artifact_file.read().decode('UTF-8')

                self.assertEqual(artifact_file_contents, 'hello world file contents')

    def test_dir_result(self):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            step_config = {
                'organization': 'test-ORG',
                'application-name': 'test-APP',
                'service-name': 'test-SERVICE',
                'version': '42.0-test'
            }

            artifact_dir_name = 'test-result-artifact-dir'
            temp_dir.makedir(artifact_dir_name)

            artifact_file_name_1 = f'{artifact_dir_name}/test-artifact1.txt'
            temp_dir.write(artifact_file_name_1, bytes('hello world 1', 'utf-8'))
            artifact_file_name_2 = f'{artifact_dir_name}/test-artifact2.txt'
            temp_dir.write(artifact_file_name_2, bytes('hello world 2', 'utf-8'))

            step_result = StepResult(
                step_name='test-step',
                sub_step_name='test-sub-step',
                sub_step_implementer_name='test-sub-step-implementer'
            )
            step_result.add_artifact(
                name='test-step-result-dir',
                value=os.path.join(temp_dir.path, artifact_dir_name)
            )
            workflow_result = WorkflowResult()
            workflow_result.add_step_result(step_result)


            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
                workflow_result=workflow_result
            )

            archive_path = step_implementer._ResultArtifactsArchive__create_archive()

            archive_zip = zipfile.ZipFile(archive_path)

            artifact_file_path_1 = f"{step_config['organization']}-" \
                f"{step_config['application-name']}-{step_config['service-name']}-" \
                f"{step_config['version']}/test-step/test-sub-step/test-step-result-dir/" \
                f"{artifact_file_name_1}"
            with archive_zip.open(artifact_file_path_1, 'r') as artifact_file:
                artifact_file_contents = artifact_file.read().decode('UTF-8')

                self.assertEqual(artifact_file_contents, 'hello world 1')

            artifact_file_path_2 = f"{step_config['organization']}-" \
                f"{step_config['application-name']}-{step_config['service-name']}-" \
                f"{step_config['version']}/test-step/test-sub-step/test-step-result-dir/" \
                f"{artifact_file_name_1}"
            with archive_zip.open(artifact_file_path_2, 'r') as artifact_file:
                artifact_file_contents = artifact_file.read().decode('UTF-8')

                self.assertEqual(artifact_file_contents, 'hello world 1')

    def test_list_result_strings(self):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            step_config = {
                'organization': 'test-ORG',
                'application-name': 'test-APP',
                'service-name': 'test-SERVICE',
                'version': '42.0-test'
            }
            step_result = StepResult(
                step_name='test-step',
                sub_step_name='test-sub-step',
                sub_step_implementer_name='test-sub-step-implementer'
            )
            step_result.add_artifact(
                name='test-step-result-str',
                value=[
                    'hello',
                    'world',
                    'foo'
                ]
            )
            workflow_result = WorkflowResult()
            workflow_result.add_step_result(step_result)
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
                workflow_result=workflow_result
            )

            archive_path = step_implementer._ResultArtifactsArchive__create_archive()

            archive_zip = zipfile.ZipFile(archive_path)

            artifact_file_path = f"{step_config['organization']}-" \
                f"{step_config['application-name']}-{step_config['service-name']}-" \
                f"{step_config['version']}/test-step/test-sub-step/test-step-result-str"
            with archive_zip.open(artifact_file_path, 'r') as artifact_file:
                artifact_file_contents = artifact_file.read().decode('UTF-8')

                self.assertEqual(
                    artifact_file_contents,
                    """[
    "hello",
    "world",
    "foo"
]"""
                )

    def test_list_result_files(self):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            step_config = {
                'organization': 'test-ORG',
                'application-name': 'test-APP',
                'service-name': 'test-SERVICE',
                'version': '42.0-test'
            }
            step_result = StepResult(
                step_name='test-step',
                sub_step_name='test-sub-step',
                sub_step_implementer_name='test-sub-step-implementer'
            )
            step_result.add_artifact(
                name='test-step-result-str',
                value=[
                    'hello',
                    'world',
                    'foo'
                ]
            )
            workflow_result = WorkflowResult()
            workflow_result.add_step_result(step_result)
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
                workflow_result=workflow_result
            )

            archive_path = step_implementer._ResultArtifactsArchive__create_archive()

            archive_zip = zipfile.ZipFile(archive_path)

            artifact_file_path = f"{step_config['organization']}-" \
                f"{step_config['application-name']}-{step_config['service-name']}-" \
                f"{step_config['version']}/test-step/test-sub-step/test-step-result-str"
            with archive_zip.open(artifact_file_path, 'r') as artifact_file:
                artifact_file_contents = artifact_file.read().decode('UTF-8')

                self.assertEqual(
                    artifact_file_contents,
                    """[
    "hello",
    "world",
    "foo"
]"""
                )

    def test_dict_result_strings(self):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            step_config = {
                'organization': 'test-ORG',
                'application-name': 'test-APP',
                'service-name': 'test-SERVICE',
                'version': '42.0-test'
            }
            step_result = StepResult(
                step_name='test-step',
                sub_step_name='test-sub-step',
                sub_step_implementer_name='test-sub-step-implementer'
            )
            step_result.add_artifact(
                name='test-step-result-str',
                value={
                    'a': 'hello',
                    'b': 'world',
                    'c': 'foo'
                }
            )
            workflow_result = WorkflowResult()
            workflow_result.add_step_result(step_result)
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
                workflow_result=workflow_result
            )

            archive_path = step_implementer._ResultArtifactsArchive__create_archive()

            archive_zip = zipfile.ZipFile(archive_path)

            artifact_file_path = f"{step_config['organization']}-" \
                f"{step_config['application-name']}-{step_config['service-name']}-" \
                f"{step_config['version']}/test-step/test-sub-step/test-step-result-str"
            with archive_zip.open(artifact_file_path, 'r') as artifact_file:
                artifact_file_contents = artifact_file.read().decode('UTF-8')

                self.assertEqual(
                    artifact_file_contents,
                    """{
    "a": "hello",
    "b": "world",
    "c": "foo"
}"""
                )

    def test_bool_result(self):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            step_config = {
                'organization': 'test-ORG',
                'application-name': 'test-APP',
                'service-name': 'test-SERVICE',
                'version': '42.0-test'
            }
            step_result = StepResult(
                step_name='test-step',
                sub_step_name='test-sub-step',
                sub_step_implementer_name='test-sub-step-implementer'
            )
            step_result.add_artifact(
                name='test-step-result-str',
                value=True
            )
            workflow_result = WorkflowResult()
            workflow_result.add_step_result(step_result)
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
                workflow_result=workflow_result
            )

            archive_path = step_implementer._ResultArtifactsArchive__create_archive()

            archive_zip = zipfile.ZipFile(archive_path)

            artifact_file_path = f"{step_config['organization']}-" \
                f"{step_config['application-name']}-{step_config['service-name']}-" \
                f"{step_config['version']}/test-step/test-sub-step/test-step-result-str"
            with archive_zip.open(artifact_file_path, 'r') as artifact_file:
                artifact_file_contents = artifact_file.read().decode('UTF-8')

                self.assertEqual(
                    artifact_file_contents,
                    'True'
                )

    def test_string_result_with_env(self):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            step_config = {
                'organization': 'test-ORG',
                'application-name': 'test-APP',
                'service-name': 'test-SERVICE',
                'version': '42.0-test'
            }
            step_result = StepResult(
                step_name='test-step',
                sub_step_name='test-sub-step',
                sub_step_implementer_name='test-sub-step-implementer',
                environment='test-env1'
            )
            step_result.add_artifact(
                name='test-step-result-str',
                value='hello world'
            )
            workflow_result = WorkflowResult()
            workflow_result.add_step_result(step_result)
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
                workflow_result=workflow_result
            )

            archive_path = step_implementer._ResultArtifactsArchive__create_archive()

            archive_zip = zipfile.ZipFile(archive_path)

            artifact_file_path = f"{step_config['organization']}-" \
                f"{step_config['application-name']}-{step_config['service-name']}-" \
                f"{step_config['version']}/test-step/test-sub-step/test-env1/test-step-result-str"
            with archive_zip.open(artifact_file_path, 'r') as artifact_file:
                artifact_file_contents = artifact_file.read().decode('UTF-8')

                self.assertEqual(artifact_file_contents, 'hello world')

    def test_file_result_with_env(self):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            step_config = {
                'organization': 'test-ORG',
                'application-name': 'test-APP',
                'service-name': 'test-SERVICE',
                'version': '42.0-test'
            }

            artifact_file_name = 'test-artifact.txt'
            temp_dir.write(artifact_file_name, bytes('hello world file contents', 'utf-8'))

            step_result = StepResult(
                step_name='test-step',
                sub_step_name='test-sub-step',
                sub_step_implementer_name='test-sub-step-implementer',
                environment='test-env1'
            )
            step_result.add_artifact(
                name='test-step-result-str',
                value=os.path.join(temp_dir.path, artifact_file_name)
            )
            workflow_result = WorkflowResult()
            workflow_result.add_step_result(step_result)


            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
                workflow_result=workflow_result
            )

            archive_path = step_implementer._ResultArtifactsArchive__create_archive()

            archive_zip = zipfile.ZipFile(archive_path)

            artifact_file_path = f"{step_config['organization']}-" \
                f"{step_config['application-name']}-{step_config['service-name']}-" \
                f"{step_config['version']}/test-step/test-sub-step/test-env1/test-step-result-str/"\
                f"{artifact_file_name}"
            with archive_zip.open(artifact_file_path, 'r') as artifact_file:
                artifact_file_contents = artifact_file.read().decode('UTF-8')

                self.assertEqual(artifact_file_contents, 'hello world file contents')



"""ConfigValueDecryptor that uses SOPS to decyrpt ConfigValues

Also See
--------
https://github.com/mozilla/sops
"""

from io import StringIO
import json
import os.path
import re
import sys
import sh

from ploigos_step_runner.config.config_value_decryptor import ConfigValueDecryptor

class SOPS(ConfigValueDecryptor):
    """ConfigValueDecryptor that uses SOPS to decyrpt ConfigValues

    Parameters
    ----------
    additional_sops_args : list
        Additional arguments to pass to the SOPS command

    Also See
    --------
    https://github.com/mozilla/sops
    """

    SOPS_ENCRYPTED_VALUE_REGEX = r'^ENC\[.*\]$'

    def __init__(self, additional_sops_args=None):
        self.__additional_sops_args = additional_sops_args

        if not self.__additional_sops_args:
            self.__additional_sops_args = []

        super().__init__()

    def can_decrypt(self, config_value):
        """Determine if a given config value can be decrypted by this decryptor.

        Parameters
        ----------
        config_value : ConfigValue
            Determine if this decryptor can decrypt this configuration value.

        Returns
        -------
        bool
            True if this ConfigValueDecryptor can decrypt the given ConfigValue
            False if this ConfigValueDecryptor can NOT decrypt the given ConfigValue.
        """
        return re.match(
            SOPS.SOPS_ENCRYPTED_VALUE_REGEX,
            str(config_value.raw_value)
        ) is not None

    def decrypt(self, config_value):
        """Decrypt the value of the given ConfigValue.

        Parameters
        ----------
        config_value : ConfigValue
            Decrypt the value of this ConfigValue.

        Returns
        -------
        obj or None
            Decrypted value of the ConfigValue
            None if this decryptor can't decrypt the given ConfigValue

        Raises
        ------
        RuntimeError
            If error attempting to run 'sops' command
        ValueError
            If given config_value#parent_source is of type string but is not a path to a file
                that exists
            If config_value#parent_source is not of type dict or str
        """
        decrypted_value = None
        sops_path = SOPS.get_sops_value_path(config_value)

        # if source is a string assume it is a file path and decrypt from that
        # else if source is a dict then dump to json and decrypt from that
        # else error
        parent_source = config_value.parent_source
        if isinstance(parent_source, str):
            # if path exists then use as file path to decrypt from
            # else error
            if os.path.exists(parent_source):
                target_file = parent_source
                stdin = None
                input_type_arg = None
            else:
                raise ValueError(
                    f"Given config value ({config_value}) parent source ({parent_source}) " +
                    "is of type (str) but is not a path to a file that exists"
                )
        elif isinstance(parent_source, dict):
            target_file = '/dev/stdin'
            stdin = json.dumps(parent_source)
            input_type_arg = '--input-type=json'
        else:
            raise ValueError(
                f"Given config value ({config_value}) parent source ({parent_source}) " +
                f"is expected to be of type dict or str but is of type: {type(parent_source)}"
            )

        try:
            # use sops to decrypt the value
            out = StringIO()
            sh.sops( # pylint: disable=no-member
                '--decrypt',
                f'--extract={sops_path}',
                input_type_arg,
                target_file,
                _in=stdin,
                _out=out,
                _err=sys.stderr,
                *self.__additional_sops_args
            )
            decrypted_value = out.getvalue()
        except sh.ErrorReturnCode as error:
            raise RuntimeError(
                f"Error invoking sops when trying to decrypt config value ({config_value}): {error}"
            ) from error

        return decrypted_value

    @staticmethod
    def get_sops_value_path(config_value):
        """Gets a stringified version of the path parts of a ConfigValue for extration
        from a SOPS encrypted file.

        Example 1
        config_value.path_parts: ['step-runner-config', 'step-foo', 0, 'config', 'test1']
        Output: '["step-runner-config"]["step-foo"][0]["config"]["test1"]'

        Returns
        -------
        str
            SOPS path Stringified version of the path parts.
        """
        path = ""

        for path_part in config_value.path_parts:
            if isinstance(path_part, str):
                path += f"[\"{path_part}\"]"
            else:
                path += f"[{path_part}]"

        return path



"""`StepImplementer` for the `validate-environment-configuration` step to take the output from
the `deploy` ArgoCD step implementer and turn it into input for the ConnfigLint implementer
for this step.

Step Configuration
------------------
Step configuration expected as input to this step.
Could come from:
* static configuration
* runtime configuration
* previous step results


| Configuration Key     | Required | Default                   | Description
|-----------------------|----------|---------------------------|---------------------------
| `argocd-deployed-manifest`   | Yes      | N/A                       | Yml file to be linted

Result Artifacts
----------------
Results artifacts output by this step.

Result Artifact Key     | Description
------------------------|------------
| `configlint-yml-file` | Yml file to be linted

"""

import os
from ploigos_step_runner import StepResult

from ploigos_step_runner import StepImplementer

DEFAULT_CONFIG = {
}

REQUIRED_CONFIG_OR_PREVIOUS_STEP_RESULT_ARTIFACT_KEYS = [
    'argocd-deployed-manifest'
]

class ConfiglintFromArgocd(StepImplementer):
    """`StepImplementer` for the `validate-environment-configuration` step to take the output from
    the `deploy` ArgoCD step implementer and turn it into input for the ConnfigLint implementer
    for this step.
    """

    @staticmethod
    def step_implementer_config_defaults():
        """
        Getter for the StepImplementer's configuration defaults.

        Returns
        -------
        dict
            Default values to use for step configuration values.

        Notes
        -----
        These are the lowest precedence configuration values.
        """
        return DEFAULT_CONFIG

    @staticmethod
    def _required_config_or_result_keys():
        """Getter for step configuration keys that are required before running the step.

        Returns
        -------
        array_list
            Array of configuration keys that are required before running the step.

        See Also
        --------
        _validate_required_config_or_previous_step_result_artifact_keys
        """
        return REQUIRED_CONFIG_OR_PREVIOUS_STEP_RESULT_ARTIFACT_KEYS

    def _run_step(self):
        """Runs the step implemented by this StepImplementer.

        Returns
        -------
        StepResult
            Object containing the dictionary results of this step.
        """
        step_result = StepResult.from_step_implementer(self)

        argocd_result_set = self.get_value('argocd-deployed-manifest')

        if not os.path.exists(argocd_result_set):
            step_result.success = False
            step_result.message = 'File specified in ' \
                                  f'argocd-deployed-manifest {argocd_result_set} not found'
            return step_result

        step_result.add_artifact(
            name='configlint-yml-path',
            value=argocd_result_set
        )
        return step_result



import os
import sh
from unittest.mock import Mock, patch
from ploigos_step_runner import StepResult, StepRunnerException, WorkflowResult, step_implementer
from ploigos_step_runner.config import Config
from ploigos_step_runner.results import StepResultArtifact
from ploigos_step_runner.step_implementers.package import NpmPackage
from pytest import raises
from tests.helpers.base_step_implementer_test_case import \
    BaseStepImplementerTestCase
from tests.helpers.test_utils import Any
from testfixtures import TempDirectory
from io import BytesIO, StringIO, TextIOWrapper

class TestNpmPackage__run_step(BaseStepImplementerTestCase):

    @patch('sh.npm', create=True) # Given a shell command, 'npm'
    @patch('os.path.exists', side_effect = lambda filename: filename == 'package.json') # Given that a file named package.json exists
    def test_run_shell_command(self, os_path_exists_mock, npm_shell_command_mock):

        # Given a working directory
        with TempDirectory() as temp_dir:
            working_dir_path = os.path.join(temp_dir.path, 'working')

            # Given an NpmPackage step implementer
            npm_test = self.create_given_step_implementer(NpmPackage, parent_work_dir_path=working_dir_path)

            # When I run the step
            npm_test.run_step()

            # Then it should run a shell command, 'npm test'
            npm_shell_command_mock.assert_any_call(
                'install',
                _out=Any(StringIO),
                _err=Any(StringIO)
            )

    @patch('sh.npm', create=True) # Given a shell command, 'npm'
    @patch('os.path.exists', side_effect = lambda filename: filename == 'package.json') # Given that a file named package.json exists
    def test_success_false_when_shell_command_fails(self, os_path_exists_mock, npm_shell_command_mock):

        # Given a working directory
        with TempDirectory() as temp_dir:
            working_dir_path = os.path.join(temp_dir.path, 'working')

            # Given the 'npm' shell command exits with an error code
            npm_shell_command_mock.side_effect = sh.ErrorReturnCode('npm', b'mock stdout', b'mock error')

            # Given an NpmPackage step implementer
            npm_test = self.create_given_step_implementer(NpmPackage, parent_work_dir_path=working_dir_path)

            # When I run the step
            step_result = npm_test.run_step()

            # Then the StepResult should have success = False
            self.assertEqual(step_result.success, False)

    @patch('sh.npm', create=True) # Given a shell command, 'npm'
    @patch('os.path.exists', side_effect = lambda filename: filename == 'package.json') # Given that a file named package.json exists
    def test_add_artifact_with_npm_output(self, os_path_exists_mock, npm_shell_command_mock):

        # Given a working directory
        with TempDirectory() as temp_dir:
            working_dir_path = os.path.join(temp_dir.path, 'working')

            # Given an NpmPackage step implementer
            npm_test = self.create_given_step_implementer(NpmPackage, parent_work_dir_path=working_dir_path)

            # When I run the step
            step_result = npm_test.run_step()

            # Then the StepResult should have an artifact with the npm output:
            artifact = step_result.get_artifact('npm-output')
            output_file_path = os.path.join(working_dir_path, 'npm_package_output.txt')
            self.assertEqual(artifact, StepResultArtifact(
                name='npm-output',
                value=output_file_path,
                description="Standard out and standard error from 'npm install'."
            ))



"""Shared utils for dealing with files.
"""

import base64
import bz2
import hashlib
import json
import os
import re
import shutil
import urllib.request
from pathlib import Path
from urllib.parse import urlparse

import yaml


def parse_yaml_or_json_file(yaml_or_json_file):
    """
    Parse YAML or JSON config files.

    Parameters
    ----------
    yaml_or_json_files : string
        List of paths to YAML or JSON files to load as a dictionary.

    Returns
    -------
    dict
        Dictionary parsed from given YAML or JSON file

    Raises
    ------
    ValueError
        If the given file can not be parsed as YAML or JSON.
    """
    parsed_file = None
    json_parse_error = None
    yaml_parse_error = None

    with open(yaml_or_json_file, 'r', encoding='utf-8') as open_yaml_or_json_file:
        file_contents = open_yaml_or_json_file.read()

    try:
        parsed_file = json.loads(file_contents)
    except ValueError as err:
        json_parse_error = err

    if not parsed_file:
        try:
            parsed_file = yaml.safe_load(file_contents)
        except (yaml.scanner.ScannerError, yaml.parser.ParserError, ValueError) as err:
            yaml_parse_error = err

    if json_parse_error and yaml_parse_error:
        raise ValueError(
            f"Error parsing file ({yaml_or_json_file}) as YAML or JSON: " +
            f"\n  JSON error: {str(json_parse_error)}" +
            f"\n  YAML error: {str(yaml_parse_error)}")

    return parsed_file

def download_source_to_destination(
    source_uri,
    destination_dir
):
    """Given a source url using a known protocol downloads the file to a given destination.

    Notes
    -----
    Known source protocols
    * file://
    * http://
    * https://

    Parameters
    ----------
    source_uri : url
        URL to a source file using a known protocol to download to destination folder
        and decompress if necessary.
    destination_dir : path
        Path to directory to download and decompress if necessary the source url to.

    Returns
    -------
    str
        Path to the downloaded file from given source.

    Raises
    ------
    RuntimeError
        If error downloading file.
    ValueError
        If source_uri does not start with file://|http://|https://
    """

    # depending on the protocol type get the source_file into the working dir
    if re.match(r'^file://|/', source_uri):
        # get the destination dir path
        if re.match(r'^file://', source_uri):
            # remove file:// and turn into an absolute path
            source_uri_abs_path = os.path.abspath(
                re.sub('^file://', '', source_uri)
            )
        else:
            source_uri_abs_path = source_uri

        # copy the file to the working dir
        source_file_name = os.path.basename(source_uri_abs_path)
        destination_path = os.path.join(destination_dir, source_file_name)
        shutil.copyfile(
            src=source_uri_abs_path,
            dst=destination_path
        )
    elif re.match(r'^http://|^https://', source_uri):
        # download the file to the working dir
        source_file_name = os.path.basename(source_uri)
        destination_path = os.path.join(destination_dir, source_file_name)

        try:
            urllib.request.urlretrieve(
                url=source_uri,
                filename=destination_path
            )
        except urllib.error.HTTPError as error:
            raise RuntimeError(f"Error downloading file ({source_uri}): {error}") from error
    else:
        # NOTE:
        #   this should NEVER happen because of the logic in
        #   _validate_required_config_or_previous_step_result_artifact_keys
        #   but rather then failing silently need to do something.
        raise ValueError(
            "Unexpected error, should have been caught by step validation."
            f" Source ({source_uri}) must start with known protocol (/|file://|http://|https://)."
        )
    return destination_path

def download_and_decompress_source_to_destination(
    source_uri,
    destination_dir
):
    """Given a source url using a known protocol downloads the file to a given destination
    and decompresses it if known compression method.

    Notes
    -----
    Known source protocols
    * file://
    * http://
    * https://

    Known compression types
    * bz2

    Parameters
    ----------
    source_uri : url
        URL to a source file using a known protocol to download to destination folder
        and decompress if necessary.
    destination_dir : path
        Path to directory to download and decompress if necessary the source url to.

    Returns
    -------
    str
        Path to the downloaded and decompressed (if needed) file from given source.

    Raises
    ------
    RuntimeError
        If error downloading file.
    ValueError
        If source_uri does not start with file://|http://|https://
    """

    destination_path = download_source_to_destination(source_uri, destination_dir)

    # if extension is .bz2, decompress, else assume file is fine as as is
    file_path, file_extension = os.path.splitext(destination_path)
    if file_extension == '.bz2':
        # NOTE: file_path is whats left after removeing .bz2 from the end
        with \
                bz2.BZ2File(destination_path) as decompressed_source, \
                open(file_path,"wb") as decompressed_destination:

            shutil.copyfileobj(decompressed_source, decompressed_destination)

            # NOTE: the compressed file was decompressed to file_path
            #       therefor that is now the actual file we want
            destination_path = file_path

    return destination_path

def create_parent_dir(file_path):
    """Helper method to create parent folder of given file if it does not exist.

    Parameters
    ----------
    file_path: str
        Absolute name of a file to create the parent folders for
    """
    parent_dir_path = os.path.dirname(file_path)
    if parent_dir_path:
        os.makedirs(parent_dir_path, exist_ok=True)

def base64_encode(file_path):
    """Given a file_path, read and encode the contents in base64
    Returns
    -------
    Base64Contents
        base64 encoded string of file contents
    """
    encoding = Path(file_path).read_text().encode('utf-8')
    return base64.b64encode(encoding).decode('utf-8')

def get_file_hash(file_path):
    """Returns file hash of given file.

    Returns
    -------
    StepResult
        Object containing the dictionary results of this step.
    """
    sha256_hash = hashlib.sha256()
    with open(file_path, "rb") as the_file:
        # Read and update hash string value in blocks of 4K
        for byte_block in iter(lambda: the_file.read(4096), b""):
            sha256_hash.update(byte_block)
    return sha256_hash.hexdigest()

def upload_file(file_path, destination_uri, username=None, password=None): # pylint: disable=too-many-locals
    """Uploads a given file to a given destination.

    Notes
    -----
    Known source protocols
    * /        - local file path
    * file://  - local file path
    * http://  - remote upload
    * https:// - remote upload

    Parameters
    ----------
    file_path : str
        Path to file to upload.
    destination_uri : str
        URI to upload file to using known protocol.
    username : str, optional
        Optional user name to use when uploading the file to an http(s) destination.
    password : str, optional
        Optional password to use when upload the file to an http(s) destination.

    Returns
    -------
    str
        If given destination is a local file path,
        then return the full path of the destination file.
        If given destination is a remote upload, return the response body as a string.

    Raises
    ------
    ValueError
        If file_path does not exist.
        If destination_uri does not start with /|file://|http://|https://
    RuntimeError
        If error uploading file.
    """
    if not os.path.exists(file_path):
        raise ValueError(f"Given file path ({file_path}) to upload does not exist.")

    upload_result = None

    if re.match(r'^file://|/', destination_uri):
        # get the destination dir path
        if re.match(r'^file://', destination_uri):
            # remove file:// and turn into an absolute path
            destination_dir_path = os.path.abspath(
                re.sub('^file://', '', destination_uri)
            )
        else:
            destination_dir_path = destination_uri

        # create the destination dir
        os.makedirs(destination_dir_path, exist_ok=True)

        # copy the file to the destination dir
        destination_file_name = os.path.basename(file_path)
        destination_path = os.path.join(destination_dir_path, destination_file_name)
        shutil.copyfile(
            src=file_path,
            dst=destination_path
        )

        upload_result = destination_path
    elif re.match(r'^http://|^https://', destination_uri):
        password_mgr = urllib.request.HTTPPasswordMgrWithDefaultRealm()

        if username:
            top_level_uri = urlparse(destination_uri).netloc
            password_mgr.add_password(None, top_level_uri, username, password)
            handler = urllib.request.HTTPBasicAuthHandler(password_mgr)
            opener = urllib.request.build_opener(handler)
        else:
            opener = urllib.request.build_opener()

        with open(file_path, 'rb') as file:
            request = urllib.request.Request(url=destination_uri, data=file.read(), method='PUT',
                                             headers={'Content-Type': 'application/octet-stream'})

            try:
                result = opener.open(request)
                response_reason = result.reason
                response_body = str(result.read(), encoding='utf8')
                response_code = result.status
                upload_result = f"status={response_code}, " \
                                f"reason={response_reason}, " \
                                f"body={response_body}"
            except urllib.error.HTTPError as error:
                raise RuntimeError(
                    f"Error uploading file ({file_path}) to destination ({destination_uri})"
                    f" with user name ({username}) and password ({password}): {error}"
                ) from error

    else:
        # NOTE:
        #   this should NEVER happen because of the logic in
        #   _validate_required_config_or_previous_step_result_artifact_keys
        #   but rather then failing silently need to do something.
        raise ValueError(
            "Unexpected error, should have been caught by step validation."
            f" Destination ({destination_uri}) must start with known protocol"
            " (/|file://|http://|https://)."
        )

    return upload_result




"""`StepImplementer` for the `uat` step using Maven by invoking the 'integration-test` maven phase.

Step Configuration
------------------
Step configuration expected as input to this step.
Could come from:
* static configuration
* runtime configuration
* previous step results

Configuration Key            | Required? | Default | Description
-----------------------------|-----------|---------|------------
`pom-file`                   | Yes       | `'pom.xml'` | pom used when executing maven.
`tls-verify`                 | No        | `True`  | Disables TLS Verification if set to False
`maven-profiles`             | No        | `[]`    | List of maven profiles to use.
`maven-no-transfer-progress` | No        | `True`  | `True` to suppress the transfer progress of packages maven downloads.
                                                     `False` to have the transfer progress printed.\
                                                      See https://maven.apache.org/ref/current/maven-embedder/cli.html
`maven-additional-arguments` | No        | `['-DskipTests']` \
                                                   | List of additional arguments to use. \
                                                     Default is because when running `integration-test` phase the `test` phase will also be run, \
                                                     so this is a "good" way to not re-run the `test` phase tests again.
`maven-servers`              | No        |         | Dictionary of dictionaries of id, username, password
`maven-repositories`         | No        |         | Dictionary of dictionaries of id, url, snapshots, releases
`maven-mirrors`              | No        |         | Dictionary of dictionaries of id, url, mirror_of
`test-reports-dir`           | No        |         | Default is to try and dynamically determine where the test reports directory is \
                                                     based on configuration in the given `pom-file`, but it is impossible task to do \
                                                     in all cases. \
                                                     So, this parameter provides a way for the user to specify where the test results \
                                                     are if our attempts at dynamically figuring it out are failing your unique pom.
`target-host-url-maven-argument-name` \
                             | Yes       |         | It is assumed that integration tests need to know a URL endpoint to run the tests against, \
                                                     but there is not standardized way for integration tests to receive that information. \
                                                     Therefor this parameter takes the name of a maven -D flag that should be set with the \
                                                     target host url to then be read by the integration tests.

Result Artifacts
----------------
Results artifacts output by this step.

Result Artifact Key | Description
--------------------|------------
`maven-output`      | Path to Stdout and Stderr from invoking Maven.
`test-report`       | Directory containing the test reports generated from running this step.
"""# pylint: disable=line-too-long

from ploigos_step_runner import StepResult
from ploigos_step_runner.exceptions import StepRunnerException
from ploigos_step_runner.step_implementers.shared import (
    MavenGeneric, MavenTestReportingMixin)

DEFAULT_CONFIG = {
    'maven-additional-arguments': ['-DskipTests']
}

REQUIRED_CONFIG_OR_PREVIOUS_STEP_RESULT_ARTIFACT_KEYS = [
    'pom-file',
    'target-host-url-maven-argument-name'
]

class MavenIntegrationTest(MavenGeneric, MavenTestReportingMixin):
    """`StepImplementer` for the `uat` step using Maven by invoking the
    'integration-test` maven phase.
    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        workflow_result,
        parent_work_dir_path,
        config,
        environment=None
    ):
        super().__init__(
            workflow_result=workflow_result,
            parent_work_dir_path=parent_work_dir_path,
            config=config,
            environment=environment,
            maven_phases_and_goals=['integration-test', 'verify']
        )


    @staticmethod
    def step_implementer_config_defaults():
        """Getter for the StepImplementer's configuration defaults.

        Returns
        -------
        dict
            Default values to use for step configuration values.

        Notes
        -----
        These are the lowest precedence configuration values.
        """
        return {**MavenGeneric.step_implementer_config_defaults(), **DEFAULT_CONFIG}

    @staticmethod
    def _required_config_or_result_keys():
        """Getter for step configuration or previous step result artifacts that are required before
        running this step.

        See Also
        --------
        _validate_required_config_or_previous_step_result_artifact_keys

        Returns
        -------
        array_list
            Array of configuration keys or previous step result artifacts
            that are required before running the step.
        """
        return REQUIRED_CONFIG_OR_PREVIOUS_STEP_RESULT_ARTIFACT_KEYS

    def _run_step(self):
        """Runs the step implemented by this StepImplementer.

        Returns
        -------
        StepResult
            Object containing the dictionary results of this step.
        """
        step_result = StepResult.from_step_implementer(self)

        # NOTE:
        #   at some point may need to do smarter logic if a deployable has more then one deployed
        #   host URL to do UAT against all of them, but for now, use first one as target of UAT
        deployed_host_urls = self.get_value('deployed-host-urls')
        if isinstance(deployed_host_urls, list):
            target_host_url = deployed_host_urls[0]
            if len(deployed_host_urls) > 1:
                step_result.message = \
                    f"Given more then one deployed host URL ({deployed_host_urls})," \
                    f" targeting first one ({target_host_url}) for user acceptance test (UAT)."
                print(step_result.message)
        elif deployed_host_urls:
            target_host_url = deployed_host_urls
        else:
            target_host_url = self.get_value('target-host-url')

        # run the tests
        print("Run user acceptance tests (UAT)")
        mvn_output_file_path = self.write_working_file('mvn_output.txt')
        try:
            # execute maven step (params come from config)
            self._run_maven_step(
                mvn_output_file_path=mvn_output_file_path,
                step_implementer_additional_arguments=[
                    f'-D{self.get_value("target-host-url-maven-argument-name")}={target_host_url}'
                ]
            )
        except StepRunnerException as error:
            step_result.success = False
            step_result.message = "Error running maven. " \
                f"More details maybe found in report artifacts: {error}"
        finally:
            step_result.add_artifact(
                description="Standard out and standard error from maven.",
                name='maven-output',
                value=mvn_output_file_path
            )

        # get test report dir
        test_report_dir = self.__get_test_report_dir()
        if test_report_dir:
            step_result.add_artifact(
                description="Test report generated when running unit tests.",
                name='test-report',
                value=test_report_dir
            )

            # gather test report evidence
            self._gather_evidence_from_test_report_directory_testsuite_elements(
                step_result=step_result,
                test_report_dir=test_report_dir
            )

        # return result
        return step_result

    def __get_test_report_dir(self):
        """Gets the test report directory.

        Search Priority:
        * values -> 'test-reports-dir'
        * pom.xml -> maven-failsafe-plugin -> reportsDirectory
        * pom.xml -> maven-surefire-plugin (for integration-test phase or goal) -> reportsDirectory

        Returns
        -------
        str
            Path to the directory containing the test reports.
        """
        # user supplied where the test reports go, just use that
        test_report_dir = self.get_value('test-reports-dir')

        # else do our best to find them
        if not test_report_dir:
            # attempt to get failsafe test report dir, if not, try for surefire
            test_report_dir = None
            try:
                test_report_dir = self._attempt_get_test_report_directory(
                    plugin_name=MavenTestReportingMixin.FAILSAFE_PLUGIN_NAME,
                    configuration_key=MavenTestReportingMixin.FAILSAFE_PLUGIN_REPORTS_DIR_CONFIG_NAME,
                    default=MavenTestReportingMixin.FAILSAFE_PLUGIN_DEFAULT_REPORTS_DIR
                )
            except StepRunnerException:
                # this means the failsafe plugin was not found, so try looking for the surefire plugin,
                # configured for this phase
                try:
                    # NOTE: when looking for surefire plugin configuration as part of the integration
                    #       test phase ensure that it is configured for correct phase, since its
                    #       default phases are for test rather then integration test.
                    test_report_dir = self._attempt_get_test_report_directory(
                        plugin_name=MavenTestReportingMixin.SUREFIRE_PLUGIN_NAME,
                        configuration_key=\
                            MavenTestReportingMixin.SUREFIRE_PLUGIN_REPORTS_DIR_CONFIG_NAME,
                        default=MavenTestReportingMixin.SUREFIRE_PLUGIN_DEFAULT_REPORTS_DIR,
                        require_phase_execution_config=True
                    )
                except StepRunnerException:
                    print(
                        'WARNING: Did not find any expected test reporting plugin'
                        f' ({MavenTestReportingMixin.FAILSAFE_PLUGIN_NAME},'
                        f' {MavenTestReportingMixin.SUREFIRE_PLUGIN_NAME})'
                        ' to read artifacts and evidence from.'
                        ' This is not wholly unexpected because there is enumerable maven plugins,'
                        ' and enumerable ways to configure them.'
                        ' Rather then relying on this step implementer to try and figure out'
                        ' where the test reports are you can configure it manually via the'
                        ' step implementer config (test-reports-dir).'
                    )

        return test_report_dir



import re
from unittest.mock import patch

from ploigos_step_runner import StepResult, StepRunner, StepRunnerException
from ploigos_step_runner.config import Config
from testfixtures import TempDirectory

from tests.helpers.base_test_case import BaseTestCase


class TestStepRunner(BaseTestCase):
    def test_init_valid_config(self):
        config = {
            'step-runner-config': {
            }
        }

        StepRunner(config)

    def test_init_invalid_config(self):
        config = {
            'blarg-config': {
            }
        }

        with self.assertRaisesRegex(
                AssertionError,
                r"Failed to add invalid config. "
                r"Missing expected top level key \(step-runner-config\): {'blarg-config': {}}"):
            StepRunner(config)

    def test_run_step_with_no_config(self):
        config = {
            'step-runner-config': {
            }
        }

        factory = StepRunner(config)

        with self.assertRaisesRegex(
           AssertionError,
               r"Can not run step \(foo\) because no step configuration provided."):
               factory.run_step('foo')

    def test_run_step_config_specified_StepImplementer_does_not_exist(self):
        config = {
            'step-runner-config': {
                'foo': [
                    {
                        'implementer': 'DoesNotExist'
                    }
                ]
            }
        }

        factory = StepRunner(config)

        with self.assertRaisesRegex(
            StepRunnerException,
                r"Could not dynamically load step \(foo\) step implementer \(DoesNotExist\)"
                r" from module \(ploigos_step_runner.step_implementers.foo\) with class name \(DoesNotExist\)"):
            factory.run_step('foo')

    def test_run_step_config_implementer_specified_and_sub_step_config_specified_StepImplementer(self):
        config = {
            'step-runner-config': {
                'foo': [
                    {
                        'implementer': 'tests.helpers.sample_step_implementers.FooStepImplementer',
                        'config': {}
                    }
                ]
            }
        }
        with TempDirectory() as temp_dir:
            factory = StepRunner(config)
            factory.run_step('foo')

    def test_run_step_config_implementer_specified_and_no_sub_step_config_specified_StepImplementer(self):
        config = {
            'step-runner-config': {
                'foo': [
                    {
                        'implementer': 'tests.helpers.sample_step_implementers.FooStepImplementer'
                    }
                ]
            }
        }
        with TempDirectory() as temp_dir:
            factory = StepRunner(config)
            factory.run_step('foo')

    def test_run_step_config_only_sub_step_and_is_dict_rather_then_array(self):
        config = {
            'step-runner-config': {
                'foo': {
                    'implementer': 'tests.helpers.sample_step_implementers.FooStepImplementer'
                }
            }
        }
        with TempDirectory() as temp_dir:
            factory = StepRunner(config)
            factory.run_step('foo')

    def test_init_with_config_object(self):
        config = {
            Config.CONFIG_KEY: {
                'step-foo': {
                    'implementer': 'foo'
                }
            }
        }
        config = Config(config)
        factory = StepRunner(config)
        self.assertEqual(factory.config, config)

    def test_init_with_dict(self):
        config = {
            Config.CONFIG_KEY: {
                'step-foo': {
                    'implementer': 'foo'
                }
            }
        }

        factory = StepRunner(config)
        sub_step_configs = factory.config.get_sub_step_configs('step-foo')
        self.assertEqual(len(sub_step_configs), 1)

    def test__get_step_implementer_class_does_not_exist_using_default_steps_module(self):
        with self.assertRaisesRegex(
                StepRunnerException,
                r"Could not dynamically load step \(foo\) step implementer \(bar\) "
                r"from module \(ploigos_step_runner.step_implementers.foo\) with class name \(bar\)"):
            StepRunner._StepRunner__get_step_implementer_class('foo', 'bar')

    def test__get_step_implementer_name_class_does_not_exist_given_module_path(self):
        with self.assertRaisesRegex(
                StepRunnerException,
                r"Could not dynamically load step \(foo\) step implementer"
                r" \(tests.helpers.sample_step_implementers.DoesNotExist\) from module \(tests.helpers.sample_step_implementers\) with class name \(DoesNotExist\)"):
            StepRunner._StepRunner__get_step_implementer_class('foo',
                                                                 'tests.helpers.sample_step_implementers.DoesNotExist')

    def test__get_step_implementer_name_class_is_not_subclass_of_StepImplementer(self):
        with self.assertRaisesRegex(
                StepRunnerException,
                re.compile(
                    r"Step \(foo\) is configured to use step implementer "
                    r"\(tests.helpers.sample_step_implementers.NotSubClassOfStepImplementer\) "
                    r"from module \(tests.helpers.sample_step_implementers\) with class name "
                    r"\(NotSubClassOfStepImplementer\), and dynamically loads as class "
                    r"\(<class 'tests.helpers.sample_step_implementers.NotSubClassOfStepImplementer'>\) "
                    r"which is not a subclass of required parent class "
                    r"\(<class 'ploigos_step_runner.step_implementer.StepImplementer'>\)."
                )
        ):
            StepRunner._StepRunner__get_step_implementer_class('foo',
                                                                 'tests.helpers.sample_step_implementers.NotSubClassOfStepImplementer')

    def test__get_step_implementer_class_exists_defaultt_steps_module(self):
        self.assertIsNotNone(
            StepRunner._StepRunner__get_step_implementer_class('container_image_static_compliance_scan', 'OpenSCAP')
        )

    def test__get_step_implementer_class_exists_include_module(self):
        self.assertIsNotNone(
            StepRunner._StepRunner__get_step_implementer_class(
                'foo',
                'tests.helpers.sample_step_implementers.FooStepImplementer'
            )
        )

    @patch('tests.helpers.sample_step_implementers.FooStepImplementer2._run_step')
    @patch('tests.helpers.sample_step_implementers.FooStepImplementer._run_step')
    def test_run_step_multiple_sub_steps_all_succeed(
        self,
        foo_step_implementer_run_step_mock,
        foo_step_implementer2_run_step_mock
    ):
        config = {
            'step-runner-config': {
                'foo': [
                    {
                        'name': 'Mock Sub Step 1',
                        'implementer': 'tests.helpers.sample_step_implementers.FooStepImplementer'
                    },
                    {
                        'name': 'Mock Sub Step 2',
                        'implementer': 'tests.helpers.sample_step_implementers.FooStepImplementer2'
                    }
                ]
            }
        }

        # mock return value
        mock_sub_step_1_result = StepResult(
            step_name='foo',
            sub_step_name='Mock Sub Step 1',
            sub_step_implementer_name='tests.helpers.sample_step_implementers.FooStepImplementer'
        )
        mock_sub_step_1_result.success = True
        mock_sub_step_2_result = StepResult(
            step_name='foo',
            sub_step_name='Mock Sub Step 2',
            sub_step_implementer_name='tests.helpers.sample_step_implementers.FooStepImplementer'
        )
        mock_sub_step_2_result.success = True

        foo_step_implementer_run_step_mock.return_value = mock_sub_step_1_result
        foo_step_implementer2_run_step_mock.return_value = mock_sub_step_2_result

        # run test
        step_runner = StepRunner(config)
        actual_success = step_runner.run_step('foo')

        # validate
        self.assertTrue(actual_success)
        foo_step_implementer_run_step_mock.assert_called_once()
        foo_step_implementer2_run_step_mock.assert_called_once()

    @patch('tests.helpers.sample_step_implementers.FooStepImplementer2._run_step')
    @patch('tests.helpers.sample_step_implementers.FooStepImplementer._run_step')
    def test_run_step_multiple_sub_steps_first_sub_step_fail(
        self,
        foo_step_implementer_run_step_mock,
        foo_step_implementer2_run_step_mock
    ):
        config = {
            'step-runner-config': {
                'foo': [
                    {
                        'name': 'Mock Sub Step 1',
                        'implementer': 'tests.helpers.sample_step_implementers.FooStepImplementer'
                    },
                    {
                        'name': 'Mock Sub Step 2',
                        'implementer': 'tests.helpers.sample_step_implementers.FooStepImplementer2'
                    }
                ]
            }
        }

        # mock return value
        mock_sub_step_1_result = StepResult(
            step_name='foo',
            sub_step_name='Mock Sub Step 1',
            sub_step_implementer_name='tests.helpers.sample_step_implementers.FooStepImplementer'
        )
        mock_sub_step_1_result.success = False
        mock_sub_step_2_result = StepResult(
            step_name='foo',
            sub_step_name='Mock Sub Step 2',
            sub_step_implementer_name='tests.helpers.sample_step_implementers.FooStepImplementer'
        )
        mock_sub_step_2_result.success = True

        foo_step_implementer_run_step_mock.return_value = mock_sub_step_1_result
        foo_step_implementer2_run_step_mock.return_value = mock_sub_step_2_result

        # run test
        step_runner = StepRunner(config)
        actual_success = step_runner.run_step('foo')

        # validate
        self.assertFalse(actual_success)
        foo_step_implementer_run_step_mock.assert_called_once()
        foo_step_implementer2_run_step_mock.assert_not_called()

    @patch('tests.helpers.sample_step_implementers.FooStepImplementer2._run_step')
    @patch('tests.helpers.sample_step_implementers.FooStepImplementer._run_step')
    def test_run_step_multiple_sub_steps_second_sub_step_fail(
        self,
        foo_step_implementer_run_step_mock,
        foo_step_implementer2_run_step_mock
    ):
        config = {
            'step-runner-config': {
                'foo': [
                    {
                        'name': 'Mock Sub Step 1',
                        'implementer': 'tests.helpers.sample_step_implementers.FooStepImplementer'
                    },
                    {
                        'name': 'Mock Sub Step 2',
                        'implementer': 'tests.helpers.sample_step_implementers.FooStepImplementer2'
                    }
                ]
            }
        }

        # mock return value
        mock_sub_step_1_result = StepResult(
            step_name='foo',
            sub_step_name='Mock Sub Step 1',
            sub_step_implementer_name='tests.helpers.sample_step_implementers.FooStepImplementer'
        )
        mock_sub_step_1_result.success = True
        mock_sub_step_2_result = StepResult(
            step_name='foo',
            sub_step_name='Mock Sub Step 2',
            sub_step_implementer_name='tests.helpers.sample_step_implementers.FooStepImplementer'
        )
        mock_sub_step_2_result.success = False

        foo_step_implementer_run_step_mock.return_value = mock_sub_step_1_result
        foo_step_implementer2_run_step_mock.return_value = mock_sub_step_2_result

        # run test
        step_runner = StepRunner(config)
        actual_success = step_runner.run_step('foo')

        # validate
        self.assertFalse(actual_success)
        foo_step_implementer_run_step_mock.assert_called_once()
        foo_step_implementer2_run_step_mock.assert_called_once()

    @patch('tests.helpers.sample_step_implementers.FooStepImplementer2._run_step')
    @patch('tests.helpers.sample_step_implementers.FooStepImplementer._run_step')
    def test_run_step_multiple_sub_steps_first_sub_step_fail_contine_sub_steps_on_failure_bool(
        self,
        foo_step_implementer_run_step_mock,
        foo_step_implementer2_run_step_mock
    ):
        config = {
            'step-runner-config': {
                'foo': [
                    {
                        'name': 'Mock Sub Step 1',
                        'continue-sub-steps-on-failure': True,
                        'implementer': 'tests.helpers.sample_step_implementers.FooStepImplementer'
                    },
                    {
                        'name': 'Mock Sub Step 2',
                        'implementer': 'tests.helpers.sample_step_implementers.FooStepImplementer2'
                    }
                ]
            }
        }

        # mock return value
        mock_sub_step_1_result = StepResult(
            step_name='foo',
            sub_step_name='Mock Sub Step 1',
            sub_step_implementer_name='tests.helpers.sample_step_implementers.FooStepImplementer'
        )
        mock_sub_step_1_result.success = False
        mock_sub_step_2_result = StepResult(
            step_name='foo',
            sub_step_name='Mock Sub Step 2',
            sub_step_implementer_name='tests.helpers.sample_step_implementers.FooStepImplementer'
        )
        mock_sub_step_2_result.success = True

        foo_step_implementer_run_step_mock.return_value = mock_sub_step_1_result
        foo_step_implementer2_run_step_mock.return_value = mock_sub_step_2_result

        # run test
        step_runner = StepRunner(config)
        actual_success = step_runner.run_step('foo')

        # validate
        self.assertFalse(actual_success)
        foo_step_implementer_run_step_mock.assert_called_once()
        foo_step_implementer2_run_step_mock.assert_called_once()

    @patch('tests.helpers.sample_step_implementers.FooStepImplementer2._run_step')
    @patch('tests.helpers.sample_step_implementers.FooStepImplementer._run_step')
    def test_run_step_multiple_sub_steps_first_sub_step_fail_contine_sub_steps_on_failure_str(
        self,
        foo_step_implementer_run_step_mock,
        foo_step_implementer2_run_step_mock
    ):
        config = {
            'step-runner-config': {
                'foo': [
                    {
                        'name': 'Mock Sub Step 1',
                        'continue-sub-steps-on-failure': 'true',
                        'implementer': 'tests.helpers.sample_step_implementers.FooStepImplementer'
                    },
                    {
                        'name': 'Mock Sub Step 2',
                        'implementer': 'tests.helpers.sample_step_implementers.FooStepImplementer2'
                    }
                ]
            }
        }

        # mock return value
        mock_sub_step_1_result = StepResult(
            step_name='foo',
            sub_step_name='Mock Sub Step 1',
            sub_step_implementer_name='tests.helpers.sample_step_implementers.FooStepImplementer'
        )
        mock_sub_step_1_result.success = False
        mock_sub_step_2_result = StepResult(
            step_name='foo',
            sub_step_name='Mock Sub Step 2',
            sub_step_implementer_name='tests.helpers.sample_step_implementers.FooStepImplementer'
        )
        mock_sub_step_2_result.success = True

        foo_step_implementer_run_step_mock.return_value = mock_sub_step_1_result
        foo_step_implementer2_run_step_mock.return_value = mock_sub_step_2_result

        # run test
        step_runner = StepRunner(config)
        actual_success = step_runner.run_step('foo')

        # validate
        self.assertFalse(actual_success)
        foo_step_implementer_run_step_mock.assert_called_once()
        foo_step_implementer2_run_step_mock.assert_called_once()



from io import StringIO
import os.path

from unittest.mock import patch

from tests.helpers.base_test_case import BaseTestCase
from tests.helpers.test_utils import Any, create_sops_side_effect

from ploigos_step_runner.config import Config, ConfigValue
from ploigos_step_runner.decryption_utils import DecryptionUtils
from ploigos_step_runner.config.decryptors.sops import SOPS

class TestConfigValue(BaseTestCase):
    def test__eq__is_equal_basic(self):
        test1 = ConfigValue('foo1', None, None)
        test2 = ConfigValue('foo1', None, None)

        self.assertEqual(test1, test2)

    def test__eq__is_equal_diff_source(self):
        test1 = ConfigValue('foo1', "does not matter for equality", None)
        test2 = ConfigValue('foo1', "really does not matter for equality", None)

        self.assertEqual(test1, test2)

    def test__eq__is_equal_diff_path_parts(self):
        test1 = ConfigValue('foo1', None, ['a','b'])
        test2 = ConfigValue('foo1', None, ['1', '2'])

        self.assertEqual(test1, test2)

    def test__eq__is_equal_diff_source_and_path_parts(self):
        test1 = ConfigValue('foo1', "does not matter for equality", ['a','b'])
        test2 = ConfigValue('foo1', "really does not matter for equality", ['1', '2'])

        self.assertEqual(test1, test2)

    def test__eq__is_not_equal_both_config_value_objects(self):
        test1 = ConfigValue('foo1', None, None)
        test2 = ConfigValue('foo2', None, None)

        self.assertNotEqual(test1, test2)

    def test__eq__is_not_equal_different_objects(self):
        test1 = ConfigValue('foo1', None, None)
        test2 = "foo1"

        self.assertNotEqual(test1, test2)

    def test__repr__(self):
        source = {
            Config.CONFIG_KEY: {
                'step-foo': [
                    {
                        'implementer': 'foo1',
                        'config': {
                            'test1': 'foo'
                        }
                    }
                ]
            }
        }

        ConfigValue.convert_leaves_to_config_values(
            values=source[Config.CONFIG_KEY],
            parent_source=source,
            path_parts=[Config.CONFIG_KEY]
        )

        self.assertEqual(
            str(source[Config.CONFIG_KEY]['step-foo'][0]['config']['test1']),
            "ConfigValue(value=foo, value_path='['step-runner-config', 'step-foo', 0, 'config', 'test1']')"
        )

    def test_convert_leaves_to_config_values_0(self):
        source = {
            Config.CONFIG_KEY: {
                'step-foo': [
                    {
                        'implementer': 'foo1',
                        'config': {
                            'test1': 'foo'
                        }
                    }
                ]
            }
        }

        ConfigValue.convert_leaves_to_config_values(
            values=source[Config.CONFIG_KEY],
            parent_source=source,
            path_parts=[Config.CONFIG_KEY]
        )

        expected = {
            Config.CONFIG_KEY: {
                'step-foo': [
                    {
                        'implementer': ConfigValue('foo1', None, None),
                        'config': {
                            'test1': ConfigValue('foo', None, None)
                        }
                    }
                ]
            }
        }

        self.assertEqual(source, expected)

    def test_convert_leaves_to_config_values_existing_ConfigValue_leaf(self):
        source = {
            Config.CONFIG_KEY: {
                'step-foo': [
                    {
                        'implementer': 'foo1',
                        'config': {
                            'test1': 'foo',
                            'test2': ConfigValue('foo1')
                        }
                    }
                ]
            }
        }

        ConfigValue.convert_leaves_to_config_values(
            values=source[Config.CONFIG_KEY],
            parent_source=source,
            path_parts=[Config.CONFIG_KEY]
        )

        expected = {
            Config.CONFIG_KEY: {
                'step-foo': [
                    {
                        'implementer': ConfigValue('foo1', None, None),
                        'config': {
                            'test1': ConfigValue('foo', None, None),
                            'test2': ConfigValue('foo1')
                        }
                    }
                ]
            }
        }

        self.assertEqual(source, expected)

    def test_convert_leaves_to_config_values_none_value_leaf(self):
        source = {
            Config.CONFIG_KEY: {
                'step-foo': [
                    {
                        'implementer': 'foo1',
                        'config': {
                            'test1': None
                        }
                    }
                ]
            }
        }

        ConfigValue.convert_leaves_to_config_values(
            values=source[Config.CONFIG_KEY],
            parent_source=source,
            path_parts=[Config.CONFIG_KEY]
        )

        expected = {
            Config.CONFIG_KEY: {
                'step-foo': [
                    {
                        'implementer': ConfigValue('foo1', None, None),
                        'config': {
                            'test1': None
                        }
                    }
                ]
            }
        }

        self.assertEqual(source, expected)

    def test_value_path_given_inital_value_path_parts(self):
        source = {
            Config.CONFIG_KEY: {
                'step-foo': [
                    {
                        'implementer': 'foo1',
                        'config': {
                            'test1': 'foo'
                        }
                    }
                ]
            }
        }

        ConfigValue.convert_leaves_to_config_values(
            values=source[Config.CONFIG_KEY],
            parent_source=source,
            path_parts=[Config.CONFIG_KEY]
        )

        self.assertEqual(
            source[Config.CONFIG_KEY]['step-foo'][0]['config']['test1'].path_parts,
            ['step-runner-config', 'step-foo', 0, 'config', 'test1'])

    def test_value_path_given_no_inital_value_path_parts(self):
        source = {
            Config.CONFIG_KEY: {
                'step-foo': [
                    {
                        'implementer': 'foo1',
                        'config': {
                            'test1': 'foo'
                        }
                    }
                ]
            }
        }

        ConfigValue.convert_leaves_to_config_values(
            values=source,
            parent_source=source
        )

        self.assertEqual(
            source[Config.CONFIG_KEY]['step-foo'][0]['config']['test1'].path_parts,
            ['step-runner-config', 'step-foo', 0, 'config', 'test1'])

    def test_convert_leaves_to_values_all_config_value_leaves(self):
        source_values = {
            Config.CONFIG_KEY: {
                'step-foo': [
                    {
                        'implementer': ConfigValue('foo1', None, None),
                        'config': {
                            'test1': ConfigValue('foo', None, None)
                        }
                    }
                ]
            }
        }

        converted = ConfigValue.convert_leaves_to_values(source_values)

        self.assertEqual(
            converted,
            {
                Config.CONFIG_KEY: {
                    'step-foo': [
                        {
                            'implementer': 'foo1',
                            'config': {
                                'test1': 'foo',
                            }
                        }
                    ]
                }
            }
        )

    def test_convert_leaves_to_values_mixed_leaves(self):
        source_values = {
            Config.CONFIG_KEY: {
                'step-foo': [
                    {
                        'implementer': ConfigValue('foo1', None, None),
                        'config': {
                            'test1': ConfigValue('foo', None, None),
                            'test': 'not a config value object'
                        }
                    }
                ]
            }
        }

        converted = ConfigValue.convert_leaves_to_values(source_values)

        self.assertEqual(
            converted,
            {
                Config.CONFIG_KEY: {
                    'step-foo': [
                        {
                            'implementer': 'foo1',
                            'config': {
                                'test1': 'foo',
                                'test': 'not a config value object'
                            }
                        }
                    ]
                }
            }
        )

    @patch('sh.sops', create=True)
    def test_value_decyrpt(self, sops_mock):
        encrypted_config_file_path = os.path.join(
            os.path.dirname(__file__),
            'decryptors',
            'files',
            'step-runner-config-secret-stuff.yml'
        )

        config_value = ConfigValue(
            value='ENC[AES256_GCM,data:UGKfnzsSrciR7GXZJhOCMmFrz3Y6V3pZsd3P,iv:yuReqA+n+rRXVHMc+2US5t7yPx54sooZSXWV4KLjDIs=,tag:jueP7/ZWLfYrEuhh+4eS8g==,type:str]',
            parent_source=encrypted_config_file_path,
            path_parts=['step-runner-config', 'global-environment-defaults', 'DEV', 'kube-api-token']
        )

        DecryptionUtils.register_config_value_decryptor(SOPS())

        sops_mock.side_effect=create_sops_side_effect('mock decrypted value')
        decrypted_value = config_value.value
        sops_mock.assert_called_once_with(
            '--decrypt',
            '--extract=["step-runner-config"]["global-environment-defaults"]["DEV"]["kube-api-token"]',
            None,
            encrypted_config_file_path,
            _in=None,
            _out=Any(StringIO),
            _err=Any(StringIO)
        )
        self.assertEqual(
            decrypted_value,
            'mock decrypted value'
        )



"""Defines a StepResultEvidence object which represents an evidence included in the StepResult
of a invocation of a StepImplementer#run.
"""

class StepResultEvidence:
    """Defines a StepResultEvidence object which represents an evidence included in the StepResult
    of a invocation of a StepImplementer#run.

    Parameters
    ----------
    name : str
        Name of the result evidence.
    value : str
        Arbitrary value of the evidence.
    description : str, optional
        Human readable description of the result evidence (defaults to empty).
    """
    def __init__(self, name, value, description=''):
        self.__name = name
        self.__value = value
        self.__description = description

    @property
    def name(self):
        """Getter for name step result evidence name.

        Returns
        -------
        str
            Step result evidence name.
        """
        return self.__name

    @property
    def value(self):
        """Getter for name step result evidence value.

        Returns
        -------
        object
            Step result evidence value.
        """
        return self.__value

    @property
    def description(self):
        """Getter for name step result evidence description.

        Returns
        -------
        str
            Step result evidence description.
        """
        return self.__description

    def as_dict(self):
        """Dictionary representation of this evidence.

        Returns
        -------
        dict
            Representation of this evidence.
        """
        return {
            'name': self.name,
            'value': self.value,
            'description': self.description
        }

    def __str__(self):
        """Get string representation of the evidence.
        """
        return str({
            'name': self.name,
            'value': self.value,
            'description': self.description
        })

    def __repr__(self):
        """Get representation of the evidence.
        """
        return "StepResultEvidence(" \
            f"name={self.name}," \
            f" value={self.value}," \
            f" description={self.description}" \
            ")"

    def __eq__(self, other):
        """StepResultEvidence is equal if all properties are equal.
        """
        return (
            isinstance(other, StepResultEvidence) and
            self.name == other.name and
            self.value == other.value and
            self.description == other.description
        )

    def __ne__(self, other):
        """StepResultEvidence is not equal if any properties are not equal.
        """
        return not self.__eq__(other)



import os.path

from testfixtures import TempDirectory

from tests.helpers.base_test_case import BaseTestCase

from ploigos_step_runner.config import Config, ConfigValue

class TestStepConfig(BaseTestCase):
    def test_parent_config(self):
        config = Config({
            Config.CONFIG_KEY: {
                'step-foo': [
                    {
                        'implementer': 'foo1',
                        'config': {
                            'test1': 'foo'
                        }
                    }
                ]

            }
        })

        step_config = config.get_step_config('step-foo')
        self.assertEqual(step_config.parent_config, config)

    def test_step_name(self):
        config = Config({
            Config.CONFIG_KEY: {
                'step-foo': [
                    {
                        'implementer': 'foo1',
                        'config': {
                            'test1': 'foo'
                        }
                    }
                ]

            }
        })

        step_config = config.get_step_config('step-foo')
        self.assertEqual(step_config.step_name, 'step-foo')

    def test_add_or_update_sub_step_config_exising_sub_step(self):
        config = Config([
            {
                Config.CONFIG_KEY: {
                    'step-foo': [
                        {
                            'implementer': 'foo1',
                            'config': {
                                'test1': 'foo'
                            }
                        }
                    ]

                }
            },
            {
                Config.CONFIG_KEY: {
                    'step-foo': [
                        {
                            'implementer': 'foo1',
                            'config': {
                                'test2': 'foo'
                            }
                        }
                    ]

                }
            },
        ])

        step_config = config.get_step_config('step-foo')
        self.assertEqual(len(step_config.sub_steps), 1)

        self.assertEqual(
            ConfigValue.convert_leaves_to_values(
                step_config.get_sub_step('foo1').sub_step_config
            ),
            {
                'test1': 'foo',
                'test2': 'foo'
            }
        )

    def test_get_sub_step_non_existing_sub_step_name(self):
        config = Config([
            {
                Config.CONFIG_KEY: {
                    'step-foo': [
                        {
                            'implementer': 'foo1',
                            'config': {
                                'test1': 'foo'
                            }
                        }
                    ]
                }
            }
        ])

        step_config = config.get_step_config('step-foo')

        self.assertIsNone(step_config.get_sub_step('does-not-exist'))



"""Defines a StepResult object which represents the results of a invocation
of a StepImplementer#run.
"""

from ploigos_step_runner import StepRunnerException
from ploigos_step_runner.results.step_result_artifact import StepResultArtifact
from ploigos_step_runner.results.step_result_evidence import StepResultEvidence


class StepResult: # pylint: disable=too-many-instance-attributes
    """Defines a StepResult object which represents the results of a invocation
    of a StepImplementer#run.

    Parameters
    ----------
    step_name : str
        Name of the step
    sub_step_name : str
        Name of the sub step
    sub_step_implementer_name : str
        Name of the sub step implementer
    environment : str
        Optional. Environment that this step result is for
        if step was run against a specific environment.
    """
    def __init__(self, step_name, sub_step_name, sub_step_implementer_name, environment=None):
        self.__step_name = step_name
        self.__sub_step_name = sub_step_name
        self.__sub_step_implementer_name = sub_step_implementer_name
        self.__environment = environment
        self.__success = True
        self.__message = ''
        self.__artifacts = {}
        self.__evidence = {}

    @classmethod
    def from_step_implementer(cls, step_implementer):
        """
        Returns
        ------
        ResultStep
        """
        return cls(
            step_name=step_implementer.step_name,
            sub_step_name=step_implementer.sub_step_name,
            sub_step_implementer_name=step_implementer.sub_step_implementer_name,
            environment=step_implementer.environment
        )

    @property
    def step_name(self):
        """
        Returns
        -------
        str
            Step name
        """
        return self.__step_name

    @property
    def sub_step_name(self):
        """
        Returns
        -------
        str
            Sub step name
        """
        return self.__sub_step_name

    @property
    def sub_step_implementer_name(self):
        """
        Returns
        -------
        str
            Sub step implementer name
        """
        return self.__sub_step_implementer_name

    @property
    def environment(self):
        """
        Returns
        -------
        str
            Environment name if this StepResult is specific to an environment,
            else None.
        """
        return self.__environment

    @property
    def artifacts(self):
        """Get the artifacts associated with this step result.

        Returns
        -------
        dict of str: StepResultArtifacts
            Key is artifact name, value is StepResultArtifact.
        """
        return self.__artifacts

    @property
    def evidence(self):
        """Get the evidence associated with this step result.

        Returns
        -------
        dict of str: StepResultEvidence
            Key is evidence name, value is StepResultEvidence.
        """
        return self.__evidence

    @property
    def artifacts_dicts(self):
        """Get the artifacts associated with this step result as dictionaries.

        Returns
        -------
        list of dict
            Each item in list a dict representation of the StepResultArtifact.
        """
        artifact_dicts = []
        for artifact in self.artifacts.values():
            artifact_dicts.append(artifact.as_dict())

        return artifact_dicts

    @property
    def evidence_dicts(self):
        """Get the evidence associated with this step result as dictionaries.

        Returns
        -------
        list of dict
            Each item in list a dict representation of the StepResultEvidence.
        """
        evidence_dicts = []
        for evidence in self.evidence.values():
            evidence_dicts.append(evidence.as_dict())

        return evidence_dicts

    def get_artifact(self, name):
        """Get artifact with given name for this StepResult.

        Parameters
        ----------
        name : str
            The name of the artifact to return.

        Returns
        -------
        StepResultArtifact
            The step result artifact with the given name for this StepResult.
        """
        return self.__artifacts.get(name)

    def get_evidence(self, name):
        """Get evidence with given name for this StepResult.

        Parameters
        ----------
        name : str
            The name of the evidence to return.

        Returns
        -------
        StepResultEvidence
            The step result evidence with the given name for this StepResult.
        """
        return self.__evidence.get(name)

    def get_artifact_value(self, name):
        """Get the value for a specified artifact.

        Parameters
        ----------
        name : str
            The name of the artifact.

        Returns
        -------
        str
            The value of the artifact.
        """
        value = None
        if self.__artifacts.get(name):
            value = self.__artifacts.get(name).value

        return value

    def get_evidence_value(self, name):
        """Get the value for a specified evidence.

        Parameters
        ----------
        name : str
            The name of the evidence.

        Returns
        -------
        str
            The value of the evidence.
        """
        value = None
        if self.evidence.get(name):
            value = self.__evidence.get(name).value

        return value

    def add_artifact(self, name, value, description=''):
        """Add an artifact to this StepResult.

        Parameters
        ----------
        name : str
            Name of the result artifact.
        value : str
            Arbitrary value of the artifact.
        description : str, optional
            Human readable description of the result artifact (defaults to empty).
        """
        if not name:
            raise StepRunnerException('Name is required to add artifact')

        # False can be the value
        if value == '' or value is None:
            raise StepRunnerException('Value is required to add artifact')

        self.__artifacts[name] = StepResultArtifact(
            name=name,
            value=value,
            description=description
        )

    def add_evidence(self, name, value, description=''):
        """Add evidence to this StepResult.

        Parameters
        ----------
        name : str
            Name of the result evidence.
        value : str
            Arbitrary value of the evidence.
        description : str, optional
            Human readable description of the result evidence (defaults to empty).
        """
        if not name:
            raise StepRunnerException('Name is required to add evidence')

        # False can be the value
        if value == '' or value is None:
            raise StepRunnerException('Value is required to add evidence')

        self.__evidence[name] = StepResultEvidence(
            name=name,
            value=value,
            description=description
        )

    @property
    def success(self):
        """
        Returns
        -------
        bool
            Success
        """
        return self.__success

    @success.setter
    def success(self, success=True):
        """Setter for success
        """
        self.__success = success

    @property
    def message(self):
        """
        Returns
        -------
        str
            Message/ error message
        """
        return self.__message

    @message.setter
    def message(self, message):
        """Setter for message
        """
        self.__message = message

    def get_sub_step_result_dict(self):
        """
        Returns
        -------
        dict
            Dictionary with the details for the sub-step.
            For example:
            {
                'sub-step-implementer-name': 'value',
                'success': Boolean,
                'message': 'value',
                'artifacts': [],
                'evidence': []
            }
        """
        result = {
            'sub-step-implementer-name': self.sub_step_implementer_name,
            'success': self.success,
            'message': self.message,
            'artifacts': self.artifacts_dicts,
            'evidence': self.evidence_dicts
        }

        return result

    def get_step_result_dict(self):
        """Get the step result dictionary

        Returns
        -------
        dict
            Results with all step result components.
            For example:
            "step-name: {
                "sub-step-name": {
                    "sub-step-implementer-name": "sub_step_implementer_name",
                    "success": True,
                    "message": "",
                    "artifacts": {
                        "name": {
                            "description": "file description",
                            "value": "step-result.txt"
                        },
                    "evidence": {
                        "name": {
                            "description": "piece of evidence",
                            "value": "evidence"
                        }
                    }
                    }
                }
            }
        """
        if self.environment:
            result = {
                self.environment: {
                    self.step_name: {
                        self.sub_step_name: self.get_sub_step_result_dict()
                    }
                }
            }
        else:
            result = {
                self.step_name: {
                    self.sub_step_name: self.get_sub_step_result_dict()
                }
            }
        return result

    def merge(self, other):
        """Merge the artifacts and evidence from another StepResult into
        this StepResult. The other StepResult must have the same step name,
        sub step name and environment.

        Parameters
        ----------
        step_result : StepResult
            The second StepResult instance to merge into this one

        Raises
        ------
        StepRunnerException if the StepResult to merge does not have a
        matching step name, sub-step name, or environment.
        """

        if not isinstance(other, StepResult):
            raise StepRunnerException('expect StepResult instance type')

        if other.step_name != self.step_name or \
            other.sub_step_name != self.sub_step_name or \
            other.environment != self.environment:
            raise StepRunnerException(
                    'Other StepResult does not have matching ' \
                    'step name, sub step name, or environment.'
                  )

        for artifact in other.artifacts.values():
            self.add_artifact(artifact.name, artifact.value, artifact.description)

        for evidence in other.evidence.values():
            self.add_evidence(evidence.name, evidence.value, evidence.description)

    def __str__(self):
        """Get string representation of the step result.
        """
        return str({
            'step-name': self.step_name,
            'sub-step-name': self.sub_step_name,
            'sub-step-implementer-name': self.sub_step_implementer_name,
            'environment': self.environment,
            'success': self.success,
            'message': self.message,
            'artifacts': self.artifacts_dicts,
            'evidence': self.evidence_dicts
        })

    def __repr__(self):
        """Get representation of the step result.
        """
        return "StepResult(" \
            f"step_name={self.step_name}," \
            f"sub_step_name={self.sub_step_name}," \
            f"sub_step_implementer_name={self.sub_step_implementer_name}," \
            f"environment={self.environment}," \
            f"success={self.success}," \
            f"message={self.message}," \
            f"artifacts={self.artifacts_dicts}" \
            f"evidence={self.evidence_dicts}" \
            ")"

    def __eq__(self, other):
        """StepResult is equal if all properties are equal.
        """
        return (
            isinstance(other, StepResult) and
            self.step_name == other.step_name and
            self.sub_step_name == other.sub_step_name and
            self.sub_step_implementer_name == other.sub_step_implementer_name and
            self.environment == other.environment and
            self.success == other.success and
            self.message == other.message and
            self.artifacts == other.artifacts and
            self.evidence == other.evidence
        )

    def __ne__(self, other):
        """StepResult is not equal if any properties are not equal.
        """
        return not self.__eq__(other)



"""Abstract parent class for StepImplementers that use Maven.

Step Configuration
------------------
Step configuration expected as input to this step.
Could come from:
* static configuration
* runtime configuration
* previous step results

Configuration Key            | Required? | Default | Description
-----------------------------|-----------|---------|-----------
`pom-file`                   | Yes       | `'pom.xml'` | pom used when executing maven.
`tls-verify`                 | No        | `True`  | Disables TLS Verification if set to False
`maven-phases-and-goals`     | Yes       |         | List of maven phases and/or goals to execute.
`maven-profiles`             | No        | `[]`    | List of maven profiles to use.
`maven-no-transfer-progress` | No        | `True`  | \
                            `True` to suppress the transfer progress of packages maven downloads.
                            `False` to have the transfer progress printed.\
                            See https://maven.apache.org/ref/current/maven-embedder/cli.html
`maven-additional-arguments` | No        | `[]`    | List of additional arguments to use.
`maven-servers`              | No        |         | Dictionary of dictionaries of id, username, password
`maven-repositories`         | No        |         | Dictionary of dictionaries of id, url, snapshots, releases
`maven-mirrors`              | No        |         | Dictionary of dictionaries of id, url, mirror_of
"""# pylint: disable=line-too-long

import os

from ploigos_step_runner import StepResult, StepRunnerException
from ploigos_step_runner.config.config_value import ConfigValue
from ploigos_step_runner.step_implementer import StepImplementer
from ploigos_step_runner.utils.maven import (generate_maven_settings,
                                            run_maven, write_effective_pom)
from ploigos_step_runner.utils.xml import get_xml_element_by_path

DEFAULT_CONFIG = {
    'pom-file': 'pom.xml',
    'tls-verify': True,
    'maven-profiles': [],
    'maven-additional-arguments': [],
    'maven-no-transfer-progress': True
}

REQUIRED_CONFIG_OR_PREVIOUS_STEP_RESULT_ARTIFACT_KEYS = [
    'pom-file',
    'maven-phases-and-goals'
]

class MavenGeneric(StepImplementer):
    """Abstract parent class for StepImplementers that use Maven.
    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        workflow_result,
        parent_work_dir_path,
        config,
        environment=None,
        maven_phases_and_goals=None
    ):
        self.__maven_settings_file = None
        self.__maven_phases_and_goals = maven_phases_and_goals

        super().__init__(
            workflow_result=workflow_result,
            parent_work_dir_path=parent_work_dir_path,
            config=config,
            environment=environment
        )

    @staticmethod
    def step_implementer_config_defaults():
        """Getter for the StepImplementer's configuration defaults.

        Returns
        -------
        dict
            Default values to use for step configuration values.

        Notes
        -----
        These are the lowest precedence configuration values.
        """
        return DEFAULT_CONFIG

    @staticmethod
    def _required_config_or_result_keys():
        """Getter for step configuration or previous step result artifacts that are required before
        running this step.

        See Also
        --------
        _validate_required_config_or_previous_step_result_artifact_keys

        Returns
        -------
        array_list
            Array of configuration keys or previous step result artifacts
            that are required before running the step.
        """
        return REQUIRED_CONFIG_OR_PREVIOUS_STEP_RESULT_ARTIFACT_KEYS

    def _validate_required_config_or_previous_step_result_artifact_keys(self):
        """Validates that the required configuration keys or previous step result artifacts
        are set and have valid values.

        Validates that:
        * required configuration is given
        * given 'pom-file' exists

        Raises
        ------
        AssertionError
            If step configuration or previous step result artifacts have invalid required values
        """
        super()._validate_required_config_or_previous_step_result_artifact_keys()

        # if pom-file has value verify file exists
        # If it doesn't have value and is required function will have already failed
        pom_file = self.get_value('pom-file')
        if pom_file is not None:
            assert os.path.exists(pom_file), \
                f'Given maven pom file (pom-file) does not exist: {pom_file}'

    @property
    def maven_phases_and_goals(self):
        """Property for getting the maven phases and goals to execute which can either come
        from field set on this class via constructor, intended for use by sub classes that want
        to hard code the phases and goals for convenience, or comes from config value
        `maven-phases-and-goals` set by the user.

        Returns
        -------
        str
            Maven phases and/or goals to execute.
        """
        maven_phases_and_goals = None
        if self.__maven_phases_and_goals:
            maven_phases_and_goals = self.__maven_phases_and_goals
        else:
            maven_phases_and_goals = self.get_value('maven-phases-and-goals')

        return maven_phases_and_goals

    @property
    def maven_settings_file(self):
        """Gets the maven settings file for this step.
        """

        if not self.__maven_settings_file:
            maven_servers = ConfigValue.convert_leaves_to_values(
                self.get_value('maven-servers')
            )
            maven_repositories = ConfigValue.convert_leaves_to_values(
                self.get_value('maven-repositories')
            )
            maven_mirrors = ConfigValue.convert_leaves_to_values(
                self.get_value('maven-mirrors')
            )

            self.__maven_settings_file = generate_maven_settings(
                working_dir=self.work_dir_path,
                maven_servers=maven_servers,
                maven_repositories=maven_repositories,
                maven_mirrors=maven_mirrors
            )

        return self.__maven_settings_file

    def _get_effective_pom(self):
        """Writes the effective pom to a file and returns the path.

        Returns
        -------
        str
            Path to the written effective pom generated from the 'pom-file' value.
        """
        effective_pom_path = os.path.join(self.work_dir_path, 'effective-pom.xml')

        if not os.path.exists(effective_pom_path):
            write_effective_pom(
                pom_file_path=self.get_value('pom-file'),
                output_path=effective_pom_path,
                profiles=self.get_value('maven-profiles')
            )

        return effective_pom_path

    def _get_effective_pom_element(self, element_path):
        """Get an XML element from the effective pom.

        Parameters
        ----------
        element_path : str
            XML path of element to to get from effective pom.

        Returns
        -------
        str
            Value of the element from the effective pom.
        """
        return get_xml_element_by_path(
            self._get_effective_pom(),
            element_path,
            default_namespace='mvn'
        )

    def _run_maven_step(
        self,
        mvn_output_file_path,
        step_implementer_additional_arguments=None
    ):
        """Runs maven using the configuration given to this step runner.

        Parameters
        ----------
        mvn_output_file_path : str
            Path to file containing the maven stdout and stderr output.
        step_implementer_additional_arguments : []
            Additional arguments hard coded by the step implementer.

        Raises
        ------
        StepRunnerException
            If maven returns a none 0 exit code.
        """

        phases_and_goals = self.maven_phases_and_goals
        pom_file = self.get_value('pom-file')
        tls_verify = self.get_value('tls-verify')
        profiles = self.get_value('maven-profiles')
        no_transfer_progress = self.get_value('maven-no-transfer-progress')

        additional_arguments = []
        if step_implementer_additional_arguments:
            additional_arguments = \
                step_implementer_additional_arguments + self.get_value('maven-additional-arguments')
        else:
            additional_arguments = self.get_value('maven-additional-arguments')

        run_maven(
            mvn_output_file_path=mvn_output_file_path,
            phases_and_goals=phases_and_goals,
            additional_arguments=additional_arguments,
            pom_file=pom_file,
            tls_verify=tls_verify,
            profiles=profiles,
            no_transfer_progress=no_transfer_progress,
            settings_file=self.maven_settings_file
        )

    def _run_step(self): # pylint: disable=too-many-locals
        """Runs the step implemented by this StepImplementer.

        Returns
        -------
        StepResult
            Object containing the dictionary results of this step.
        """
        step_result = StepResult.from_step_implementer(self)

        # package the artifacts
        mvn_output_file_path = self.write_working_file('mvn_output.txt')
        try:
            # execute maven step (params come from config)
            self._run_maven_step(
                mvn_output_file_path=mvn_output_file_path
            )
        except StepRunnerException as error:
            step_result.success = False
            step_result.message = "Error running maven. " \
                f"More details maybe found in 'maven-output' report artifact: {error}"
        finally:
            step_result.add_artifact(
                description="Standard out and standard error from maven.",
                name='maven-output',
                value=mvn_output_file_path
            )

        return step_result



# pylint: disable=missing-module-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=missing-function-docstring
import os
import re
from io import IOBase
from unittest.mock import patch

import sh
from testfixtures import TempDirectory
from tests.helpers.base_step_implementer_test_case import \
    BaseStepImplementerTestCase
from tests.helpers.test_utils import Any
from ploigos_step_runner.exceptions import StepRunnerException
from ploigos_step_runner.step_implementers.tag_source import Git
from ploigos_step_runner import StepResult


class TestStepImplementerTagSourceGit(BaseStepImplementerTestCase):
    def create_step_implementer(
        self,
        workflow_result=None,
        step_config={},
        parent_work_dir_path=''
    ):
        return self.create_given_step_implementer(
            step_implementer=Git,
            step_config=step_config,
            step_name='tag-source',
            implementer='Git',
            workflow_result=workflow_result,
            parent_work_dir_path=parent_work_dir_path
        )

# TESTS FOR configuration checks
    def test_step_implementer_config_defaults(self):
        defaults = Git.step_implementer_config_defaults()
        expected_defaults = {}
        self.assertEqual(defaults, expected_defaults)

    def test__required_config_or_result_keys(self):
        required_keys = Git._required_config_or_result_keys()
        expected_required_keys = []
        self.assertEqual(required_keys, expected_required_keys)

    def test__validate_required_config_or_previous_step_result_artifact_keys_valid(self):
         with TempDirectory() as test_dir:
            parent_work_dir_path = os.path.join(test_dir.path, 'working')

            step_config = {
                'git-username': 'git-username',
                'git-password': 'git-password'
            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
            )

            step_implementer._validate_required_config_or_previous_step_result_artifact_keys()

    def test__validate_required_config_or_previous_step_result_artifact_keys_invalid_missing_git_username(self):
         with TempDirectory() as test_dir:
            parent_work_dir_path = os.path.join(test_dir.path, 'working')

            step_config = {
                'git-password': 'git-password'
            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
            )

            with self.assertRaisesRegex(
                StepRunnerException,
                r"Either 'git-username' or 'git-password 'is not set. Neither or both must be set."
            ):
                step_implementer._validate_required_config_or_previous_step_result_artifact_keys()

    def test__validate_required_config_or_previous_step_result_artifact_keys_invalid_missing_git_password(self):
         with TempDirectory() as test_dir:
            parent_work_dir_path = os.path.join(test_dir.path, 'working')

            step_config = {
                'git-username': 'git-username'
            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
            )

            with self.assertRaisesRegex(
                StepRunnerException,
                r"Either 'git-username' or 'git-password 'is not set. Neither or both must be set."
            ):
                step_implementer._validate_required_config_or_previous_step_result_artifact_keys()


# TESTS FOR _run_step

    @patch.object(Git, '_Git__get_tag')
    @patch.object(Git, '_Git__git_url')
    @patch.object(Git, '_Git__git_tag')
    @patch.object(Git, '_Git__git_push')
    def test_run_step_pass(self, git_push_mock, git_tag_mock, git_url_mock, get_tag_mock):
        with TempDirectory() as temp_dir:
            tag = '1.0+69442c8'
            url = 'git@github.com:ploigos/ploigos-step-runner.git'
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')

            step_config = {
                'url': url,
                'git-username': 'git-username',
                'git-password': 'git-password'
            }

            artifact_config = {
                'version': {'description': '', 'value': tag},
                'container-image-version': {'description': '', 'value': tag}
            }

            workflow_result = self.setup_previous_result(parent_work_dir_path, artifact_config)

            step_implementer = self.create_step_implementer(
                step_config=step_config,
                workflow_result=workflow_result,
                parent_work_dir_path=parent_work_dir_path,
            )

            def get_tag_side_effect():
                return tag
            get_tag_mock.side_effect = get_tag_side_effect

            def git_url_side_effect():
                return url
            git_url_mock.side_effect = git_url_side_effect


            result = step_implementer._run_step()

            expected_step_result = StepResult(step_name='tag-source', sub_step_name='Git', sub_step_implementer_name='Git')
            expected_step_result.add_artifact(name='tag', value=tag)

            # verifying all mocks were called
            get_tag_mock.assert_called_once_with()
            git_tag_mock.assert_called_once_with(tag)
            git_url_mock.assert_called_once_with()
            git_push_mock.assert_called_once_with(None)

            self.assertEqual(result, expected_step_result)

    @patch.object(Git, '_Git__get_tag')
    @patch.object(Git, '_Git__git_url')
    @patch.object(Git, '_Git__git_tag')
    @patch.object(Git, '_Git__git_push')
    def test_run_step_pass_http(self, git_push_mock, git_tag_mock, git_url_mock, get_tag_mock):
        with TempDirectory() as temp_dir:
            tag = '1.0+69442c8'
            url = 'http://git.ploigos.xyz/ploigos-references/ploigos-reference-app-quarkus-rest-json.git'
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')

            artifact_config = {
                'container-image-version': {'description': '', 'value': tag}
            }
            workflow_result = self.setup_previous_result(parent_work_dir_path, artifact_config)

            step_config = {
                'url': url,
                'git-username': 'git-username',
                'git-password': 'git-password'
            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                workflow_result=workflow_result,
                parent_work_dir_path=parent_work_dir_path
            )

            def get_tag_side_effect():
                return tag
            get_tag_mock.side_effect = get_tag_side_effect

            def git_url_side_effect():
                return url
            git_url_mock.side_effect = git_url_side_effect


            result = step_implementer._run_step()

            expected_step_result = StepResult(step_name='tag-source', sub_step_name='Git', sub_step_implementer_name='Git')
            expected_step_result.add_artifact(name='tag', value=tag)

            # verifying all mocks were called
            get_tag_mock.assert_called_once_with()
            git_tag_mock.assert_called_once_with(tag)
            git_url_mock.assert_called_once_with()
            git_push_mock.assert_called_once_with('http://git-username:git-password@' + url[7:])

            self.assertEqual(result, expected_step_result)

    @patch.object(Git, '_Git__get_tag')
    @patch.object(Git, '_Git__git_url')
    @patch.object(Git, '_Git__git_tag')
    @patch.object(Git, '_Git__git_push')
    def test_run_step_pass_https(self, git_push_mock, git_tag_mock, git_url_mock, get_tag_mock):
        with TempDirectory() as temp_dir:
            tag = '1.0+69442c8'
            url = 'https://git.ploigos.xyz/ploigos-references/ploigos-reference-app-quarkus-rest-json.git'
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')

            artifact_config = {
                'version': {'description': '', 'value': tag},
                'container-image-version': {'description': '', 'value': tag}
            }
            workflow_result = self.setup_previous_result(parent_work_dir_path, artifact_config)

            step_config = {
                'url': url,
                'git-username': 'git-username',
                'git-password': 'git-password'
            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                workflow_result=workflow_result,
                parent_work_dir_path=parent_work_dir_path
            )

            def get_tag_side_effect():
                return tag
            get_tag_mock.side_effect = get_tag_side_effect

            def git_url_side_effect():
                return url
            git_url_mock.side_effect = git_url_side_effect


            result = step_implementer._run_step()

            expected_step_result = StepResult(step_name='tag-source', sub_step_name='Git', sub_step_implementer_name='Git')
            expected_step_result.add_artifact(name='tag', value=tag)

            # verifying all mocks were called
            get_tag_mock.assert_called_once_with()
            git_tag_mock.assert_called_once_with(tag)
            git_url_mock.assert_called_once_with()
            git_push_mock.assert_called_once_with('https://git-username:git-password@' + url[8:])

            self.assertEqual(result, expected_step_result)

    @patch.object(Git, '_Git__get_tag')
    @patch.object(Git, '_Git__git_url')
    @patch.object(Git, '_Git__git_tag')
    @patch.object(Git, '_Git__git_push')
    def test_run_step_pass_no_username_password(self, git_push_mock, git_tag_mock, git_url_mock, get_tag_mock):
        with TempDirectory() as temp_dir:
            tag = '1.0+69442c8'
            url = 'git@github.com:ploigos/ploigos-step-runner.git'
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')

            artifact_config = {
                'version': {'description': '', 'value': tag},
                'container-image-version': {'description': '', 'value': tag}
            }
            workflow_result = self.setup_previous_result(parent_work_dir_path, artifact_config)

            step_config = {
                'url': url
            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                workflow_result=workflow_result,
                parent_work_dir_path=parent_work_dir_path
            )

            def get_tag_side_effect():
                return tag
            get_tag_mock.side_effect = get_tag_side_effect

            def git_url_side_effect():
                return url
            git_url_mock.side_effect = git_url_side_effect


            result = step_implementer._run_step()

            expected_step_result = StepResult(step_name='tag-source', sub_step_name='Git', sub_step_implementer_name='Git')
            expected_step_result.add_artifact(name='tag', value=tag)

            # verifying all mocks were called
            get_tag_mock.assert_called_once_with()
            git_tag_mock.assert_called_once_with(tag)
            git_url_mock.assert_called_once_with()
            git_push_mock.assert_called_once_with(None)

            self.assertEqual(result, expected_step_result)

    @patch.object(Git, '_Git__get_tag')
    @patch.object(Git, '_Git__git_url')
    @patch.object(Git, '_Git__git_tag')
    @patch.object(Git, '_Git__git_push')
    def test_run_step_fail_http_no_username_password(self, git_push_mock, git_tag_mock, git_url_mock, get_tag_mock):
        with TempDirectory() as temp_dir:
            tag = '1.0+69442c8'
            url = 'http://git.ploigos.xyz/ploigos-references/ploigos-reference-app-quarkus-rest-json.git'
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')

            artifact_config = {
                'version': {'description': '', 'value': tag},
                'container-image-version': {'description': '', 'value': tag}
            }
            workflow_result = self.setup_previous_result(parent_work_dir_path, artifact_config)

            step_config = {
                'url': url
            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                workflow_result=workflow_result,
                parent_work_dir_path=parent_work_dir_path
            )

            def get_tag_side_effect():
                return tag
            get_tag_mock.side_effect = get_tag_side_effect

            def git_url_side_effect():
                return url
            git_url_mock.side_effect = git_url_side_effect


            result = step_implementer._run_step()

            expected_step_result = StepResult(step_name='tag-source', sub_step_name='Git', sub_step_implementer_name='Git')
            expected_step_result.success = False
            expected_step_result.message = 'For a http:// git url, you need to also provide username/password pair'

            # verifying all mocks were called
            get_tag_mock.assert_called_once_with()
            git_tag_mock.assert_called_once_with(tag)
            git_url_mock.assert_called_once_with()

            self.assertEqual(result, expected_step_result)

    @patch.object(Git, '_Git__get_tag')
    @patch.object(Git, '_Git__git_url')
    @patch.object(Git, '_Git__git_tag')
    @patch.object(Git, '_Git__git_push')
    def test_run_step_fail_https_no_username_password(self, git_push_mock, git_tag_mock, git_url_mock, get_tag_mock):
        with TempDirectory() as temp_dir:
            tag = '1.0+69442c8'
            url = 'https://git.ploigos.xyz/ploigos-references/ploigos-reference-app-quarkus-rest-json.git'
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')

            artifact_config = {
                'version': {'description': '', 'value': tag},
                'container-image-version': {'description': '', 'value': tag}
            }
            workflow_result = self.setup_previous_result(parent_work_dir_path, artifact_config)

            step_config = {
                'url': url
            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                workflow_result=workflow_result,
                parent_work_dir_path=parent_work_dir_path
            )

            def get_tag_side_effect():
                return tag
            get_tag_mock.side_effect = get_tag_side_effect

            def git_url_side_effect():
                return url
            git_url_mock.side_effect = git_url_side_effect

            result = step_implementer._run_step()

            expected_step_result = StepResult(step_name='tag-source', sub_step_name='Git', sub_step_implementer_name='Git')
            expected_step_result.success = False
            expected_step_result.message = 'For a https:// git url, you need to also provide username/password pair'

            # verifying all mocks were called
            get_tag_mock.assert_called_once_with()
            git_tag_mock.assert_called_once_with(tag)
            git_url_mock.assert_called_once_with()

            self.assertEqual(result, expected_step_result)

    @patch.object(Git, '_Git__get_tag')
    @patch.object(Git, '_Git__git_url')
    @patch.object(Git, '_Git__git_tag')
    @patch.object(Git, '_Git__git_push')
    def test_run_step_error_git_tag(self, git_push_mock, git_tag_mock, git_url_mock, get_tag_mock):
        with TempDirectory() as temp_dir:
            tag = '1.0+69442c8'
            url = 'git@github.com:ploigos/ploigos-step-runner.git'
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')

            artifact_config = {
                'version': {'description': '', 'value': tag},
                'container-image-version': {'description': '', 'value': tag}
            }
            workflow_result = self.setup_previous_result(parent_work_dir_path, artifact_config)

            step_config = {
                'url': url,
                'git-username': 'git-username',
                'git-password': 'git-password'
            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                workflow_result=workflow_result,
                parent_work_dir_path=parent_work_dir_path
            )

            def get_tag_side_effect():
                return tag
            get_tag_mock.side_effect = get_tag_side_effect

            def git_url_side_effect():
                return url
            git_url_mock.side_effect = git_url_side_effect

            # this is the test
            git_tag_mock.side_effect = StepRunnerException('mock error')
            result = step_implementer._run_step()

            # verify results
            expected_step_result = StepResult(
                step_name='tag-source',
                sub_step_name='Git',
                sub_step_implementer_name='Git'
            )
            expected_step_result.add_artifact(name='tag', value=tag)
            expected_step_result.success = False
            expected_step_result.message = "mock error"

            # verifying correct mocks were called
            git_tag_mock.assert_called_once_with(tag)

            self.assertEqual(result, expected_step_result)

    @patch.object(Git, '_Git__get_tag')
    @patch.object(Git, '_Git__git_url')
    @patch.object(Git, '_Git__git_tag')
    @patch.object(Git, '_Git__git_push')
    def test_run_step_error_git_url(self, git_push_mock, git_tag_mock, git_url_mock, get_tag_mock):
        with TempDirectory() as temp_dir:
            tag = '1.0+69442c8'
            url = 'git@github.com:ploigos/ploigos-step-runner.git'
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')

            artifact_config = {
                'version': {'description': '', 'value': tag},
                'container-image-version': {'description': '', 'value': tag}
            }
            workflow_result = self.setup_previous_result(parent_work_dir_path, artifact_config)

            step_config = {
                'url': url,
                'git-username': 'git-username',
                'git-password': 'git-password'
            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                workflow_result=workflow_result,
                parent_work_dir_path=parent_work_dir_path
            )

            def get_tag_side_effect():
                return tag
            get_tag_mock.side_effect = get_tag_side_effect

            # this is the test here
            git_url_mock.side_effect = StepRunnerException('mock error')
            result = step_implementer._run_step()

            # verify test results
            expected_step_result = StepResult(
                step_name='tag-source',
                sub_step_name='Git',
                sub_step_implementer_name='Git'
            )
            expected_step_result.add_artifact(name='tag', value=tag)
            expected_step_result.success = False
            expected_step_result.message = "mock error"

            # verifying correct mocks were called
            git_tag_mock.assert_called_once_with(tag)
            git_url_mock.assert_called_once_with()

            self.assertEqual(result, expected_step_result)

    @patch.object(Git, '_Git__get_tag')
    @patch.object(Git, '_Git__git_url')
    @patch.object(Git, '_Git__git_tag')
    @patch.object(Git, '_Git__git_push')
    def test_run_step_error_git_push(self, git_push_mock, git_tag_mock, git_url_mock, get_tag_mock):
        with TempDirectory() as temp_dir:
            tag = '1.0+69442c8'
            url = 'git@github.com:ploigos/ploigos-step-runner.git'
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')

            artifact_config = {
                'version': {'description': '', 'value': tag},
                'container-image-version': {'description': '', 'value': tag}
            }
            workflow_result = self.setup_previous_result(parent_work_dir_path, artifact_config)

            step_config = {
                'url': url,
                'git-username': 'git-username',
                'git-password': 'git-password'
            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                workflow_result=workflow_result,
                parent_work_dir_path=parent_work_dir_path
            )

            def get_tag_side_effect():
                return tag
            get_tag_mock.side_effect = get_tag_side_effect

            def git_url_side_effect():
                return url
            git_url_mock.side_effect = git_url_side_effect

            # this is the test here
            git_push_mock.side_effect = StepRunnerException('mock error')
            result = step_implementer._run_step()

            # verify the test results
            expected_step_result = StepResult(
                step_name='tag-source',
                sub_step_name='Git',
                sub_step_implementer_name='Git'
            )
            expected_step_result.add_artifact(name='tag', value=tag)
            expected_step_result.success = False
            expected_step_result.message = "mock error"

            # verifying all mocks were called
            get_tag_mock.assert_called_once_with()
            git_tag_mock.assert_called_once_with(tag)
            git_url_mock.assert_called_once_with()
            git_push_mock.assert_called_once_with(None)

            self.assertEqual(result, expected_step_result)

# TEST FOR __get_tag

    @patch.object(Git, '_Git__git_url')
    @patch.object(Git, '_Git__git_tag')
    @patch.object(Git, '_Git__git_push')
    def test__get_tag_no_version(self, git_push_mock, git_tag_mock, git_url_mock):
        with TempDirectory() as temp_dir:
            tag = 'latest'
            url = 'git@github.com:ploigos/ploigos-step-runner.git'
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')

            artifact_config = {
                'container-image-version': {'description': '', 'value': tag}
            }
            workflow_result = self.setup_previous_result(parent_work_dir_path, artifact_config)

            step_config = {
                'url': url
            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                workflow_result=workflow_result,
                parent_work_dir_path=parent_work_dir_path
            )

            def git_url_side_effect():
                return url
            git_url_mock.side_effect = git_url_side_effect


            result = step_implementer._run_step()

            expected_step_result = StepResult(step_name='tag-source', sub_step_name='Git', sub_step_implementer_name='Git')
            expected_step_result.add_artifact(name='tag', value=tag)

            # verifying all mocks were called
            git_tag_mock.assert_called_once_with(tag)
            git_url_mock.assert_called_once_with()
            git_push_mock.assert_called_once_with(None)

            self.assertEqual(result, expected_step_result)

# TESTS FOR __git_url

    def create_git_config_side_effect(self, mock_remote_origin_url):
        def git_config_side_effect(*args, **kwargs):
            if (args[0] == '--get' and args[1] == 'remote.origin.url'):
                kwargs['_out'].write(mock_remote_origin_url)

        return git_config_side_effect

    @patch('sh.git', create=True)
    def test__git_url_url_from_git_config(self, git_mock):
        remote_origin_url = "https://does.not.matter.xyz/foo.git"

        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')

            artifact_config = {
                'container-image-version': {'description': '', 'value': '1.0-69442c8'}
            }
            workflow_result = self.setup_previous_result(parent_work_dir_path, artifact_config)

            step_config = {}
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                workflow_result=workflow_result,
                parent_work_dir_path=parent_work_dir_path
            )

            git_mock.config.side_effect = TestStepImplementerTagSourceGit.\
                create_git_config_side_effect(self, remote_origin_url)

            url = step_implementer._Git__git_url()

            self.assertEqual(url, remote_origin_url)

    @patch('sh.git', create=True)
    def test__git_url_url_from_git_config_error(self, git_mock):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')

            artifact_config = {
                'container-image-version': {'description': '', 'value': '1.0-69442c8'}
            }
            workflow_result = self.setup_previous_result(parent_work_dir_path, artifact_config)

            step_config = {}
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                workflow_result=workflow_result,
                parent_work_dir_path=parent_work_dir_path
            )

            git_mock.config.side_effect = sh.ErrorReturnCode('git', b'mock out', b'mock error')

            with self.assertRaisesRegex(
                StepRunnerException,
                "Error invoking git config --get remote.origin.url:"
            ):
                step_implementer._Git__git_url()

    @patch('sh.git', create=True)
    def test__git_url_from_step_config(self, git_mock):
        remote_origin_url = "https://does.not.matter.xyz/foo.git"

        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')

            artifact_config = {
                'container-image-version': {'description': '', 'value': '1.0-69442c8'}
            }
            workflow_result = self.setup_previous_result(parent_work_dir_path, artifact_config)

            step_config = {
                'url': remote_origin_url
            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                workflow_result=workflow_result,
                parent_work_dir_path=parent_work_dir_path
            )

            git_mock.assert_not_called()

            url = step_implementer._Git__git_url()

            self.assertEqual(url, remote_origin_url)

# TESTS FOR __git_tag

    @patch('sh.git', create=True)
    def test__git_tag_success(self, git_mock):
        git_tag_value = "1.0+69442c8"

        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')

            artifact_config = {
                'container-image-version': {'description': '', 'value': '1.0-69442c8'}
            }
            workflow_result = self.setup_previous_result(parent_work_dir_path, artifact_config)

            step_config = {}
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                workflow_result=workflow_result,
                parent_work_dir_path=parent_work_dir_path
            )

            step_implementer._Git__git_tag(git_tag_value)

            git_mock.tag.assert_called_once_with(
                git_tag_value,
                '-f',
                _out=Any(IOBase),
                _err=Any(IOBase),
                _tee='err'
            )

    @patch('sh.git', create=True)
    def test__git_tag_error(self, git_mock):
        git_tag_value = "1.0+69442c8"

        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')

            artifact_config = {
                'container-image-version': {'description': '', 'value': '1.0-69442c8'}
            }
            workflow_result = self.setup_previous_result(parent_work_dir_path, artifact_config)

            step_config = {}
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                workflow_result=workflow_result,
                parent_work_dir_path=parent_work_dir_path
            )

            git_mock.tag.side_effect = sh.ErrorReturnCode('git', b'mock out', b'mock error')

            with self.assertRaisesRegex(
                StepRunnerException,
                "Error pushing git tag"
            ):
                step_implementer._Git__git_tag(git_tag_value)

# TESTS FOR __git_push

    @patch('sh.git', create=True)
    def test__git_push_with_url_success(self, git_mock):
        url = 'www.xyz.com'

        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')

            artifact_config = {
                'container-image-version': {'description': '', 'value': '1.0-69442c8'}
            }
            workflow_result = self.setup_previous_result(parent_work_dir_path, artifact_config)

            step_config = {}
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                workflow_result=workflow_result,
                parent_work_dir_path=parent_work_dir_path
            )

            step_implementer._Git__git_push(url)

            git_mock.push.assert_called_once_with(
                url,
                '--tag',
                _out=Any(IOBase),
                _err=Any(IOBase),
                _tee='err'
            )

    @patch('sh.git', create=True)
    def test__git_push_no_url_success(self, git_mock):

        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')

            artifact_config = {
                'container-image-version': {'description': '', 'value': '1.0-69442c8'}
            }
            workflow_result = self.setup_previous_result(parent_work_dir_path, artifact_config)

            step_config = {}
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                workflow_result=workflow_result,
                parent_work_dir_path=parent_work_dir_path
            )

            step_implementer._Git__git_push(None)

            git_mock.push.assert_called_once_with(
                '--tag',
                _out=Any(IOBase),
                _err=Any(IOBase),
                _tee='err'
            )

    @patch('sh.git', create=True)
    def test__git_push_error(self, git_mock):
        url = 'www.xyz.com'

        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')

            artifact_config = {
                'container-image-version': {'description': '', 'value': '1.0-69442c8'}
            }
            workflow_result = self.setup_previous_result(parent_work_dir_path, artifact_config)

            step_config = {}
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                workflow_result=workflow_result,
                parent_work_dir_path=parent_work_dir_path
            )

            git_mock.push.side_effect = sh.ErrorReturnCode('git', b'mock out', b'mock error')

            with self.assertRaisesRegex(
                StepRunnerException,
                "Error invoking git push"
            ):
                step_implementer._Git__git_push(url)
